; |=======================================================================|
; |                                                                       |
; |                                OCaPIC                                 |
; |                                                                       |
; |            Benoit Vaugon, Université Pierre et Marie Curie            |
; |                                                                       |
; |    Ce fichier est distribué sous les termes de la licence CeCILL-B    |
; |    décrite dans le fichier ../../LICENCE.                             |
; |                                                                       |
; |=======================================================================|

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;             FAILURES              ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

M_RAISE_FROM_GLOBAL macro G
		M_CHECK_UNFULL_HEAP 0x1
		lfsr	FSR0, (STACK_ANCHOR - 0x1 - (0x2 * G))
		clrf	POSTINC1	; write tag
		movlw	0x1
		movwf	POSTINC1	; write size
		movff	FSR1L, ACCUL	; ACCU <- exn
		movff	FSR1H, ACCUH
		movff	POSTINC0, POSTINC1 ; block[0] <- global(W)
		movff	POSTINC0, POSTINC1
		goto	_RAISE
	endm

M_WRITE_BYTE macro C
		movlw	C
		movwf	POSTINC1
	endm

caml_raise_out_of_memory:
	movlw	low OOM_B0L
	movwf	ACCUL
	movlw	high OOM_B0H
	movwf	ACCUH
	goto	_RAISE

caml_raise_division_by_0:
	M_RAISE_FROM_GLOBAL DIVISION_BY_0_IND
	
caml_raise_stack_overflow:
	M_RAISE_FROM_GLOBAL STACK_OVERFLOW_IND


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;               HEAP                ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	;; check allocation in heap
	;; start gc if necessary
	;; raise exception if memory overflow
	;; give its argument (the size of block) in TABLAT
_CHECK_UNFULL_HEAP_GEN:
	movf	CUR_HEAP_END, W
	cpfslt	FSR1H
	bra	_CHECK_UNFULL_HEAP_GEN_L0
	addlw	-0x1
	cpfseq	FSR1H
	return

	;; if FSR1H = CUR_HEAP_END - 1
	movf	FSR1L, W
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L1
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L2
	return

	;; if FSR1H = CUR_HEAP_END
_CHECK_UNFULL_HEAP_GEN_L0:
	movf	FSR1L, W
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L3
_CHECK_UNFULL_HEAP_GEN_L1:
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L3
_CHECK_UNFULL_HEAP_GEN_L2:
	addlw	0x2
	bc	_CHECK_UNFULL_HEAP_GEN_L3
	return

_CHECK_UNFULL_HEAP_GEN_L3:
	;; execute GC
	rcall	caml_gc_exec

	;; check a new time
	movf	CUR_HEAP_END, W
	cpfslt	FSR1H
	bra	_CHECK_UNFULL_HEAP_GEN_L4
	addlw	-0x1
	cpfseq	FSR1H
	return

	;; if FSR1H = CUR_HEAP_END - 1
	movf	FSR1L, W
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L5
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L6
	return

	;; if FSR1H = CUR_HEAP_END
_CHECK_UNFULL_HEAP_GEN_L4:
	movf	FSR1L, W
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L7
_CHECK_UNFULL_HEAP_GEN_L5:
	addwf	TABLAT, W
	bc	_CHECK_UNFULL_HEAP_GEN_L7
_CHECK_UNFULL_HEAP_GEN_L6:
	addlw	0x2
	btfss	STATUS, C
	return
_CHECK_UNFULL_HEAP_GEN_L7:
	goto	caml_raise_out_of_memory


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;          ALLOC DUMMY              ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

caml_alloc_dummy:
caml_alloc_dummy_float:
	rrcf	ACCUH, W		; STATUS.C ignored
	rrcf	ACCUL, W		; TABLAT <- size
	bz	caml_alloc_dummy_atom	; size = 0 ?
	movwf	TABLAT
	call	_CHECK_UNFULL_HEAP_GEN
	clrf	POSTINC1		; write tag
	movf	TABLAT, W		; write size
	movwf	POSTINC1
	addwf	FSR1L, F		; FSR1 += 2 * size
	btfsc	STATUS, C
	incf	FSR1H, F
	addwf	FSR1L, F
	btfsc	STATUS, C
	incf	FSR1H, F
	return
caml_alloc_dummy_atom:
	movlw	low (ATOM0_SIZE + 1)
	movwf	ACCUL
	movlw	high (ATOM0_SIZE + 1)
	movwf	ACCUH
	return
	
caml_update_dummy:
	movff	FSR1L, TMP_REG_1 	; save FSR1
	movff	FSR1H, TMP_REG_2
	movff	ACCUL, FSR1L		; FSR1 <- dummy
	movff	ACCUH, FSR1H
	movsf	[0x1], FSR0L		; FSR0 <- newval
	movsf	[0x2], FSR0H
	subfsr	FSR0, 0x2
	subfsr	FSR1, 0x2
	movff	POSTINC0, POSTINC1	; copy tag
	movf	POSTINC0, W		; read size
	bz	caml_update_dummy_end
	addfsr	FSR1, 0x1
caml_update_dummy_loop:
	movff	POSTINC0, POSTINC1 	; loop size times
	movff	POSTINC0, POSTINC1
	decfsz	WREG, F
	bra	caml_update_dummy_loop
caml_update_dummy_end:
	movff	TMP_REG_1, FSR1L 	; restore FSR1
	movff	TMP_REG_2, FSR1H
	return

	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;                 GC                ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

caml_gc_exec:
;;; PHASE 1: copy all roots if necessary

	;; exchange heap anchors and ends
	;; FSR1 <- new current heap anchor
	clrf	FSR1L
	movf	OTH_HEAP_ANCHOR, W
	movwf	FSR1H
	movff	CUR_HEAP_ANCHOR, OTH_HEAP_ANCHOR
	movwf	CUR_HEAP_ANCHOR
	movf	OTH_HEAP_END, W
	movff	CUR_HEAP_END, OTH_HEAP_END
	movwf	CUR_HEAP_END

	;; save stack top in GC_TMP_REG_2:1
	;; FSR0 <- stack top
	movf	FSR2L, W
	movwf	GC_TMP_REG_1
	movwf	FSR0L
	movf	FSR2H, W
	movwf	GC_TMP_REG_2
	movwf	FSR0H

	;; check and copy stack elements
	addfsr	FSR0, 0x1
	bra	caml_gc_stack_loop_test
caml_gc_stack_loop:
	rcall	caml_check_and_copy
caml_gc_stack_loop_test:
	movlw	(low (STACK_ANCHOR + 0x1))
	cpfseq	FSR0L
	bra	caml_gc_stack_loop
	movlw	(high (STACK_ANCHOR + 0x1))
	cpfseq	FSR0H
	bra	caml_gc_stack_loop
	
	;; check and copy environment and accu
	rcall	caml_check_and_copy
	rcall	caml_check_and_copy

;;; PHASE 2: sweep new heap and upgrade it

	;; FSR0 <- new heap anchor
	clrf	FSR0L
	movff	CUR_HEAP_ANCHOR, FSR0H

	;; loop on new heap
	bra	caml_gc_heap_loop_test
caml_gc_heap_loop:
	movlw	NO_SCAN_TAG
	cpfsgt	POSTINC0	; read tag
	bra	caml_gc_scan_block
	movf	POSTINC0, W	; skip block ; read size
	addwf	FSR0L, F	; FSR0 += 2 * size
	btfsc	STATUS, C
	incf	FSR0H, F
	addwf	FSR0L, F
	btfsc	STATUS, C
	incf	FSR0H, F
	bra	caml_gc_heap_loop_test
caml_gc_scan_block:
	movff	POSTINC0, GC_TMP_REG_3	; GC_TMP_REG_3 <- size
caml_gc_loop_block:
	rcall	caml_check_and_copy
	decfsz	GC_TMP_REG_3, F		; loop size times
	bra	caml_gc_loop_block
caml_gc_heap_loop_test:			; if FSR0 = FSR1 then break else loop
	movf	FSR0L, W
	cpfseq	FSR1L
	bra	caml_gc_heap_loop
	movf	FSR0H, W
	cpfseq	FSR1H
	bra	caml_gc_heap_loop
	
	;; restore stack top
	movff	GC_TMP_REG_1, FSR2L
	movff	GC_TMP_REG_2, FSR2H
	;; increment counter
	infsnz	GC_COUNTER_L, F
	incf	GC_COUNTER_H, F
	return


caml_check_and_copy:
;;; check *FSR0
;;; if it is a pointer in the old heap then
;;;        if it has not been copied then copy;
;;;        update *FSR0;
;;; increment FSR0 by 2
	movf	POSTINC0, W		; W <- (*FSR0++)L
	btfsc	WREG, 0			; isint ?
	bra	caml_copy_skip_1 	; yes -> skip
	movwf	FSR2L			; FSR2L <- low pointer
	movf	POSTDEC0, W		; W <- (*FSR0--)H
	movwf	FSR2H			; FSR2H <- high pointer
	subwf	OTH_HEAP_END, W		; in old heap ?
	bnc	caml_copy_skip_2 	; no -> skip
	addlw	-HEAP_SIZE
	bc	caml_copy_skip_2 	; no -> skip
	subfsr	FSR2, 0x2		; FSR2 -= 2
	movlw	INFIX_TAG		; is infix block ?
	cpfseq	[0x0]
	bra	caml_copy_block		; no -> just copy
	movf	[0x1], W		; yes -> shift FSR2
	movwf	GC_TMP_REG_5		; save offset in GC_TMP_REG_5
	subwf	FSR2L, F		; FSR2 -= 2 * offset
	btfss	STATUS, C
	decf	FSR2H, F
	subwf	FSR2L, F
	btfss	STATUS, C
	decf	FSR2H, F
	rcall	caml_copy_block		; copy the encapsulating block
	movf	GC_TMP_REG_5, W		; shift result by 2 * offset
	subfsr	FSR0, 0x2
	addwf	POSTINC0, F
	btfsc	STATUS, C
	incf	INDF0, F
	subfsr	FSR0, 0x1
	addwf	POSTINC0, F
	btfsc	STATUS, C
	incf	INDF0, F
	addfsr	FSR0, 0x1
	return
caml_copy_block:
	movf	[0x1], W		; is already copied ? read block size
	bnz	caml_copy_not_copied	; no -> copy
	movf	[0x2], W		; yes -> just update *FSR0
	movwf	POSTINC0
	movf	[0x3], W
	movwf	POSTINC0
	return
caml_copy_not_copied:		   	; copy block
	movff	POSTINC2, POSTINC1	; copy tag ; FSR2 ++ ; FSR1 ++
	movf	INDF2, W		; read size
	clrf	POSTINC2		; clear size ; FSR2 ++
	movwf	POSTINC1		; copy size
	movwf	GC_TMP_REG_4		; GC_TMP_REG_4 <- size
	movf	FSR1L, W		; *FSR0 <- FSR1 ; *FSR2 <- FSR1
	movff	INDF2, POSTINC1		; *FSR1 <- *FSR2
	movwf	POSTINC0		; FSR0 += 2 ; FSR1 += 2 ; FSR2 += 2
	movwf	POSTINC2
	movf	FSR1H, W		; ok because address are even
	movff	INDF2, POSTINC1
	movwf	POSTINC0
	movwf	POSTINC2
	decf	GC_TMP_REG_4, F		; GC_TMP_REG_4 --
	bz	caml_copy_skip_0	; if size = 1 then skip loop
caml_copy_block_loop:
	movff	POSTINC2, POSTINC1
	movff	POSTINC2, POSTINC1
	decfsz	GC_TMP_REG_4, F		; loop size - 1 times
	bra	caml_copy_block_loop
	return
caml_copy_skip_2:
	movf	POSTINC0, F		; FSR0 ++
caml_copy_skip_1:
	movf	POSTINC0, F		; FSR0 ++
caml_copy_skip_0:
	return
