; |=======================================================================|
; |                                                                       |
; |                                OCaPIC                                 |
; |                                                                       |
; |            Benoit Vaugon, Université Pierre et Marie Curie            |
; |                                                                       |
; |    Ce fichier est distribué sous les termes de la licence CeCILL-B    |
; |    décrite dans le fichier ../../LICENCE.                             |
; |                                                                       |
; |=======================================================================|

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;             STRINGS               ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_blit_string
caml_blit_string:
	;; ACCU = src
	;; [0x2]:[0x1] = srcoff
	;; [0x4]:[0x3] = dst
	;; [0x6]:[0x5] = dstoff
	;; [0x8]:[0x7] = len
	movff	FSR1L, TMP_REG_1 ; save FSR1 in TMP_REG_2:1
	movff	FSR1H, TMP_REG_2
	bcf	STATUS, C	; FSR0 <- src + srcoff
	rrcf	[0x2], F
	rrcf	[0x1], W
	addwf	ACCUL, W
	movwf	FSR0L
	movf	[0x2], W
	addwfc	ACCUH, W
	movwf	FSR0H
	bcf	STATUS, C	; FSR1 <- dst + dstoff
	rrcf	[0x6], F
	rrcf	[0x5], W
	addwf	[0x3], W
	movwf	FSR1L
	movf	[0x6], W
	addwfc	[0x4], W
	movwf	FSR1H
	bcf	STATUS, C	; [0x8]:[0x7] <- len
	rrcf	[0x8], F
	rrcf	[0x7], F
	bz	caml_blit_string_loop_test
caml_blit_string_loop:
	movff	POSTINC0, POSTINC1
	decfsz	[0x7], F
	bra	caml_blit_string_loop
caml_blit_string_loop_test:
	decf	[0x8], F
	bz	caml_blit_string_loop
	movff	TMP_REG_1, FSR1L ; restore FSR1
	movff	TMP_REG_2, FSR1H
	M_CONST	0		; return unit
	return
#endif

#ifdef caml_useprim_caml_ml_string_length
caml_ml_string_length:
	;; ACCU = str
	movff	ACCUL, FSR0L	; FSR0 <- ACCU
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1	; W <- block_size
	movf	POSTINC0, W
	addwf	FSR0L, F	; FSR0 += 2 * block_size - 1 ; W --
	btfsc	STATUS, C
	incf	FSR0H, F
	addlw	-0x1
	addwf	FSR0L, F
	btfsc	STATUS, C
	incf	FSR0H, F
	movwf	ACCUL		; ACCU <- int_val(2 * W)
	clrf	ACCUH
	bcf	STATUS, C
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bsf	STATUS, C
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	btfss	INDF0, 0	; if INDF0.0 = 0 then ACCU ++
	bsf	ACCUL, 1
	return
#endif

#ifdef caml_useprim_caml_fill_string
caml_fill_string:
	;; ACCU = str
	;; [0x2]:[0x1] = start
	;; [0x4]:[0x3] = len
	;; [0x6]:[0x5] = c
	bcf	STATUS, C	; FSR0 <- str + start
	rrcf	[0x2], F
	rrcf	[0x1], W
	addwf	ACCUL, W
	movwf	FSR0L
	movf	[0x2], W
	addwfc	ACCUH, W
	movwf	FSR0H
	rrcf	[0x6], W	; [0x5] <- c
	rrcf	[0x5], F
	bcf	STATUS, C	; [0x4]:[0x3] <- len
	rrcf	[0x4], F
	rrcf	[0x3], F
	bz	caml_fill_string_miloop
caml_fill_string_loop:
	movf	[0x5], W
	movwf	POSTINC0
	decfsz	[0x3], F
	bra	caml_fill_string_loop
caml_fill_string_miloop:
	decf	[0x4], F
	bz	caml_fill_string_loop
	M_CONST	0		; return unit
	return
#endif

#ifdef caml_useprim_caml_is_printable
caml_is_printable:
	M_CONST	1
	return
#endif

#ifdef caml_useprim_caml_create_string
#ifndef caml_useprim_caml_raise_ia_string_create
#define caml_useprim_caml_raise_ia_string_create
#endif
caml_create_string:
	;; ACCU = len
	rrcf	ACCUH, W	; STATUS.C ignored
	movwf	TMP_REG_1
	rrcf	ACCUL, W
	movwf	TABLAT
	movwf	TMP_REG_2
	rrcf	TMP_REG_1, W	; STATUS.C ignored
	andlw	0x3F
	bnz	caml_raise_ia_string_create
	rrcf	TABLAT, F
	incf	TABLAT, F
	bz	caml_raise_ia_string_create
	call	_CHECK_UNFULL_HEAP_GEN
	movlw	STRING_TAG	; write tag
	movwf	POSTINC1
	movf	TABLAT, W	; write size
	movwf	POSTINC1
	movff	FSR1L, ACCUL	; ACCU <- string
	movff	FSR1H, ACCUH
	addlw	-0x1		; FSR1 += 2 * size - 2
	addwf	FSR1L, F
	btfsc	STATUS, C
	incf	FSR1H, F
	addwf	FSR1L, F
	btfsc	STATUS, C
	incf	FSR1H, F
	clrf	POSTINC1       	; write 0
	movlw	0x0		; if odd size then write 0 else write 1
	btfss	TMP_REG_2, 0
	movlw	0x1
	movwf	POSTINC1
	return
#endif

#ifdef caml_useprim_caml_string_get
#ifndef caml_useprim_caml_raise_ia_index_out_of_bounds_string
#define caml_useprim_caml_raise_ia_index_out_of_bounds_string
#endif
caml_string_get:
	;; ACCU = str
	;; [0x2]:[0x1] = ind
	movff	ACCUL, FSR0L	; FSR0 <- @size
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1
	bcf	STATUS, C	; [0x2]:[0x1] <- valint(ind)
	rrcf	[0x2], F
	rrcf	[0x1], F
	bcf	STATUS, C	; check indH
	rrcf	[0x2], W	; C <- LSB(indH)
	bnz	caml_raise_ia_index_out_of_bounds_string
	rrcf	[0x1], W	; W <- ind / 2
	cpfsgt	INDF0		; size > ind / 2 ?
	bra	caml_raise_ia_index_out_of_bounds_string ; no -> error
	addlw	0x1		; W <- ind / 2 + 1
	cpfsgt	POSTINC0	; size = ind / 2 + 1 ?
	bra	caml_string_get_last_char ; yes -> goto last_char
	movf	[0x1], W	; no -> ok ; FSR0 += size
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	clrf	ACCUH		; read char
	bsf	STATUS, C
	rlcf	INDF0, W
	rlcf	ACCUH, F
	movwf	ACCUL
	return
caml_string_get_last_char:
	btfsc	[0x1], 0	; parity of ind ?
	bra	caml_raise_ia_index_out_of_bounds_string ; odd -> error
	movf	[0x1], W	; FSR0 <- @last byte (0 or 1)
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	addfsr	FSR0, 0x1
	movf	POSTDEC0, W	; test last byte
	bnz	caml_raise_ia_index_out_of_bounds_string ; if 1 -> raise
	clrf	ACCUH		; if 0 -> ok, read char
	bsf	STATUS, C
	rlcf	INDF0, W
	rlcf	ACCUH, F
	movwf	ACCUL
	return
#endif

#ifdef caml_useprim_caml_string_set
#ifndef caml_useprim_caml_raise_ia_index_out_of_bounds_string
#define caml_useprim_caml_raise_ia_index_out_of_bounds_string
#endif
caml_string_set:
	;; ACCU = str
	;; [0x2]:[0x1] = ind
	;; [0x4]:[0x3] = c
	movff	ACCUL, FSR0L	; FSR0 <- @size
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1
	bcf	STATUS, C	; [0x2]:[0x1] <- valint(ind)
	rrcf	[0x2], F
	rrcf	[0x1], F
	bcf	STATUS, C	; check indH
	rrcf	[0x2], W	; C <- LSB(indH)
	bnz	caml_raise_ia_index_out_of_bounds_string
	rrcf	[0x1], W	; W <- ind / 2
	cpfsgt	INDF0		; size > ind / 2 ?
	bra	caml_raise_ia_index_out_of_bounds_string ; no -> error
	addlw	0x1		; W <- ind / 2 + 1
	cpfsgt	POSTINC0	; size = ind / 2 + 1 ?
	bra	caml_string_set_last_char ; yes -> goto last_char
	movf	[0x1], W	; no -> ok ; FSR0 += size
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	rrcf	[0x4], W	; W <- c ; STATUS.C ignored
	rrcf	[0x3], W
	movwf	INDF0		; write char
	M_CONST	0		; return ()
	return
caml_string_set_last_char:
	btfsc	[0x1], 0	; parity of ind ?
	bra	caml_raise_ia_index_out_of_bounds_string ; odd -> error
	movf	[0x1], W	; FSR0 <- @last byte (0 or 1)
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	addfsr	FSR0, 0x1
	movf	POSTDEC0, W	; test last byte
	bnz	caml_raise_ia_index_out_of_bounds_string ; if 1 -> raise
	rrcf	[0x4], W	; W <- c ; STATUS.C ignored
	rrcf	[0x3], W
	movwf	INDF0		; write char
	M_CONST	0		; return ()
	return
#endif

#ifdef caml_useprim_caml_raise_ia_index_out_of_bounds_string
#ifndef caml_useprim_caml_raise_ia_index_out_of_bounds
#define caml_useprim_caml_raise_ia_index_out_of_bounds
#endif
caml_raise_ia_index_out_of_bounds_string:
	movlw	0x1			; erase ACCU and stack[0]
	movwf	ACCUL
	movwf	[0x1]
	clrf	ACCUH
	clrf	[0x2]
	bra	caml_raise_ia_index_out_of_bounds
#endif

#ifdef caml_useprim_caml_raise_ia_string_create
#ifndef caml_useprim_caml_raise_ia
#define caml_useprim_caml_raise_ia
#endif
caml_raise_ia_string_create:
	M_CHECK_UNFULL_HEAP 0xA		; check allocation of string and exn
	M_WRITE_BYTE STRING_TAG		; write string tag
	M_WRITE_BYTE 0x7		; write string size
	movff	FSR1L, TMP_REG_1	; mem string addr
	movff	FSR1H, TMP_REG_2
	M_WRITE_BYTE 'S'		; write string content
	M_WRITE_BYTE 't'
	M_WRITE_BYTE 'r'
	M_WRITE_BYTE 'i'
	M_WRITE_BYTE 'n'
	M_WRITE_BYTE 'g'
	M_WRITE_BYTE '.'
	M_WRITE_BYTE 'c'
	M_WRITE_BYTE 'r'
	M_WRITE_BYTE 'e'
	M_WRITE_BYTE 'a'
	M_WRITE_BYTE 't'
	M_WRITE_BYTE 'e'
	M_WRITE_BYTE 0x0		; write string \0
	goto	caml_raise_ia
#endif


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;              ARRAYS               ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; should compress float array
#ifdef caml_useprim_caml_make_array
caml_make_array:
	return
#endif

#ifdef caml_useprim_caml_make_vect
#ifndef caml_useprim_caml_raise_ia_array_make
#define caml_useprim_caml_raise_ia_array_make
#endif
caml_make_vect:
	;; ACCU = size
	;; [0x2]:[0x1] = val
	bcf	STATUS, C
	rrcf	ACCUH, W
	bnz	caml_raise_ia_array_make
	rrcf	ACCUL, W
	bz	caml_make_vect_atom
	movwf	TABLAT
	call	_CHECK_UNFULL_HEAP_GEN
	clrf	POSTINC1		; write tag (0)
	movff	TABLAT, POSTINC1	; write size
	movff	FSR1L, ACCUL
	movff	FSR1H, ACCUH
caml_make_vect_loop:
	movf	[0x1], W
	movwf	POSTINC1
	movf	[0x2], W
	movwf	POSTINC1
	decfsz	TABLAT, F
	bra	caml_make_vect_loop
	return
caml_make_vect_atom:
	movlw	low (ATOM0_SIZE + 1)
	movwf	ACCUL
	movlw	high (ATOM0_SIZE + 1)
	movwf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_raise_ia_array_make
#ifndef caml_useprim_caml_raise_ia
#define caml_useprim_caml_raise_ia
#endif
caml_raise_ia_array_make:
	M_CHECK_UNFULL_HEAP 0x9		; check allocation of string and exn
	M_WRITE_BYTE STRING_TAG		; write string tag
	M_WRITE_BYTE 0x6		; write string size
	movff	FSR1L, TMP_REG_1	; mem string addr
	movff	FSR1H, TMP_REG_2
	M_WRITE_BYTE 'A'		; write string content
	M_WRITE_BYTE 'r'
	M_WRITE_BYTE 'r'
	M_WRITE_BYTE 'a'
	M_WRITE_BYTE 'y'
	M_WRITE_BYTE '.'
	M_WRITE_BYTE 'm'
	M_WRITE_BYTE 'a'
	M_WRITE_BYTE 'k'
	M_WRITE_BYTE 'e'
	M_WRITE_BYTE 0x0		; write string \0\1
	M_WRITE_BYTE 0x1
	goto	caml_raise_ia
#endif

#ifdef caml_useprim_caml_array_unsafe_get
#ifndef caml_useprim_caml_array_unsafe_get_addr
#define caml_useprim_caml_array_unsafe_get_addr
#endif
#ifndef caml_useprim_caml_array_unsafe_get_float
#define caml_useprim_caml_array_unsafe_get_float
#endif
caml_array_unsafe_get:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	movff	ACCUL, FSR0L		; FSR0 <- ACCU - 2
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x2
	movlw	DOUBLE_ARRAY_TAG 	; tag = double[] tag ?
	cpfseq	POSTINC0
	bra	caml_array_unsafe_get_addr ; no
	bra	caml_array_unsafe_get_float ; yes
#endif

#ifdef caml_useprim_caml_array_get
#ifndef caml_useprim_caml_array_get_addr
#define caml_useprim_caml_array_get_addr
#endif
#ifndef caml_useprim_caml_array_get_float
#define caml_useprim_caml_array_get_float
#endif
caml_array_get:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	movff	ACCUL, FSR0L		; FSR0 <- ACCU - 2
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x2
	movlw	DOUBLE_ARRAY_TAG	; tag = double[] tag ?
	cpfseq	POSTINC0
	bra	caml_array_get_addr 	; no
	bra	caml_array_get_float	; yes
#endif

#ifdef caml_useprim_caml_array_unsafe_set
#ifndef caml_useprim_caml_array_unsafe_set_addr
#define caml_useprim_caml_array_unsafe_set_addr
#endif
#ifndef caml_useprim_caml_array_unsafe_set_float
#define caml_useprim_caml_array_unsafe_set_float
#endif
caml_array_unsafe_set:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	;; [0x4]:[0x3] = val
	movff	ACCUL, FSR0L		; FSR0 <- ACCU - 2
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x2
	movlw	DOUBLE_ARRAY_TAG	; tag = double[] tag ?
	cpfseq	POSTINC0
	bra	caml_array_unsafe_set_addr ; no
	bra	caml_array_unsafe_set_float ; yes
#endif

#ifdef caml_useprim_caml_array_set
#ifndef caml_useprim_caml_array_set_addr
#define caml_useprim_caml_array_set_addr
#endif
#ifndef caml_useprim_caml_array_set_float
#define caml_useprim_caml_array_set_float
#endif
caml_array_set:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	;; [0x4]:[0x3] = val
	movff	ACCUL, FSR0L		; FSR0 <- ACCU - 2
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x2
	movlw	DOUBLE_ARRAY_TAG	; tag = double[] tag ?
	cpfseq	POSTINC0
	bra	caml_array_set_addr 	; no
	bra	caml_array_set_float	; yes
#endif

#ifdef caml_useprim_caml_array_unsafe_get_addr
caml_array_unsafe_get_addr:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	decf	[0x1], W	; FSR0 <- tbl + ind
	addwf	ACCUL, W
	movwf	FSR0L
	movf	ACCUH, W
	addwfc	[0x2], W
	movwf	FSR0H
	movff	POSTINC0, ACCUL	; ACCU <- *FSR0
	movff	INDF0, ACCUH
	return
#endif

#ifdef caml_useprim_caml_array_unsafe_set_addr
caml_array_unsafe_set_addr:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	;; [0x4]:[0x3] = val
	decf	[0x1], W	; FSR0 <- tbl + ind
	addwf	ACCUL, W
	movwf	FSR0L
	movf	ACCUH, W
	addwfc	[0x2], W
	movwf	FSR0H
	movf	[0x3], W	; *FSR0 <- val
	movwf	POSTINC0
	movf	[0x4], W
	movwf	INDF0
	M_CONST	0		; return ()
	return
#endif

#ifdef caml_useprim_caml_array_unsafe_get_float
caml_array_unsafe_get_float:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	M_CHECK_UNFULL_HEAP 0x1
	movff	ACCUL, FSR0L	; FSR0 <- ACCU + 2 * ind
	movff	ACCUH, FSR0H
	decf	[0x1], W
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	movlw	DOUBLE_TAG	; write tag
	movwf	POSTINC1
	movlw	0x1		; write size
	movwf	POSTINC1
	movff	FSR1L, ACCUL	; ACCU <- @float
	movff	FSR1H, ACCUH
	movff	POSTINC0, POSTINC1 ; copy float
	movff	POSTINC0, POSTINC1
	return
#endif

#ifdef caml_useprim_caml_array_unsafe_set_float
caml_array_unsafe_set_float:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	;; [0x4]:[0x3] = val
	movsf	[0x3], FSR0L		; TMP_REG_2:1 <- *val
	movsf	[0x4], FSR0H
	movff	POSTINC0, TMP_REG_1
	movff	INDF0, TMP_REG_2
	movff	ACCUL, FSR0L		; FSR0 <- ACCU + 2 * ind
	movff	ACCUH, FSR0H
	decf	[0x1], W
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	movff	TMP_REG_1, POSTINC0 	; write the float
	movff	TMP_REG_2, POSTINC0
	M_CONST	0			; return ()
	return
#endif

#ifdef caml_useprim_caml_array_get_addr
#ifndef caml_useprim_caml_raise_ia_index_out_of_bounds
#define caml_useprim_caml_raise_ia_index_out_of_bounds
#endif
caml_array_get_addr:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	movff	ACCUL, FSR0L	; FSR0 <- @size
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1
	bcf	STATUS, C	; check indH
	rrcf	[0x2], W	; C <- LSB(indH)
	bnz	caml_raise_ia_index_out_of_bounds
	rrcf	[0x1], W	; W <- ind
	cpfsgt	POSTINC0	; size > ind ? FSR0++
	bra	caml_raise_ia_index_out_of_bounds ; no -> raise
	decf	[0x1], W	; FSR0 += ind
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	movff	POSTINC0, ACCUL	; ACCU <- *FSR0
	movff	INDF0, ACCUH
	return
#endif

#ifdef caml_useprim_caml_array_get_float
#ifndef caml_useprim_caml_raise_ia_index_out_of_bounds
#define caml_useprim_caml_raise_ia_index_out_of_bounds
#endif
caml_array_get_float:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	M_CHECK_UNFULL_HEAP 0x1
	movff	ACCUL, FSR0L	; FSR0 <- ACCU
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1
	bcf	STATUS, C	; check indH
	rrcf	[0x2], W	; C <- LSB(indH)
	bnz	caml_raise_ia_index_out_of_bounds
	rrcf	[0x1], W	; W <- ind
	cpfsgt	POSTINC0	; size > ind ? FSR0++
	bra	caml_raise_ia_index_out_of_bounds ; no -> raise
	decf	[0x1], W	; FSR0 += 2 * ind
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	movlw	DOUBLE_TAG	; write tag
	movwf	POSTINC1
	movlw	0x1		; write size
	movwf	POSTINC1
	movff	FSR1L, ACCUL	; ACCU <- @float
	movff	FSR1H, ACCUH
	movff	POSTINC0, POSTINC1 ; copy float
	movff	POSTINC0, POSTINC1
	return
#endif

#ifdef caml_useprim_caml_array_set_addr
#ifndef caml_useprim_caml_raise_ia_index_out_of_bounds
#define caml_useprim_caml_raise_ia_index_out_of_bounds
#endif
caml_array_set_addr:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	;; [0x4]:[0x3] = val
	movff	ACCUL, FSR0L	; FSR0 <- @size
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1
	bcf	STATUS, C	; check indH
	rrcf	[0x2], W	; C <- LSB(indH)
	bnz	caml_raise_ia_index_out_of_bounds
	rrcf	[0x1], W	; W <- ind
	cpfsgt	POSTINC0	; size > ind ? FSR0++
	bra	caml_raise_ia_index_out_of_bounds ; no -> raise
	decf	[0x1], W	; FSR0 += ind
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	movf	[0x3], W	; *FSR0 <- val
	movwf	POSTINC0
	movf	[0x4], W
	movwf	INDF0
	M_CONST	0		; return ()
	return
#endif

#ifdef caml_useprim_caml_array_set_float
#ifndef caml_useprim_caml_raise_ia_index_out_of_bounds
#define caml_useprim_caml_raise_ia_index_out_of_bounds
#endif
caml_array_set_float:
	;; ACCU = tbl
	;; [0x2]:[0x1] = ind
	;; [0x4]:[0x3] = val
	movsf	[0x3], FSR0L		; TMP_REG_2:1 <- *val
	movsf	[0x4], FSR0H
	movff	POSTINC0, TMP_REG_1
	movff	INDF0, TMP_REG_2
	movff	ACCUL, FSR0L		; FSR0 <- ACCU
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1
	bcf	STATUS, C	; check indH
	rrcf	[0x2], W	; C <- LSB(indH)
	bnz	caml_raise_ia_index_out_of_bounds
	rrcf	[0x1], W	; W <- ind
	cpfsgt	POSTINC0	; size > ind ? FSR0++
	bra	caml_raise_ia_index_out_of_bounds ; no -> raise
	decf	[0x1], W	; FSR0 += ind
	addwf	FSR0L, F
	movf	[0x2], W
	addwfc	FSR0H, F
	movff	TMP_REG_1, POSTINC0 	; write the float
	movff	TMP_REG_2, POSTINC0
	M_CONST	0			; return ()
	return
#endif

#ifdef caml_useprim_caml_raise_ia_index_out_of_bounds
#ifndef caml_useprim_caml_raise_ia
#define caml_useprim_caml_raise_ia
#endif
caml_raise_ia_index_out_of_bounds:
	M_CHECK_UNFULL_HEAP 0xD		; check allocation of string and exn
	M_WRITE_BYTE STRING_TAG		; write string tag
	M_WRITE_BYTE 0xA		; write string size
	movff	FSR1L, TMP_REG_1	; mem string addr
	movff	FSR1H, TMP_REG_2
	M_WRITE_BYTE 'i'		; write string content
	M_WRITE_BYTE 'n'
	M_WRITE_BYTE 'd'
	M_WRITE_BYTE 'e'
	M_WRITE_BYTE 'x'
	M_WRITE_BYTE ' '
	M_WRITE_BYTE 'o'
	M_WRITE_BYTE 'u'
	M_WRITE_BYTE 't'
	M_WRITE_BYTE ' '
	M_WRITE_BYTE 'o'
	M_WRITE_BYTE 'f'
	M_WRITE_BYTE ' '
	M_WRITE_BYTE 'b'
	M_WRITE_BYTE 'o'
	M_WRITE_BYTE 'u'
	M_WRITE_BYTE 'n'
	M_WRITE_BYTE 'd'
	M_WRITE_BYTE 's'
	M_WRITE_BYTE 0x0		; write string \0
	goto	caml_raise_ia
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;              HASH                 ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_hash_univ_param
caml_hash_univ_param:
	;; ACCU = count
	;; [0x2]:[0x1] = limit
	;; [0x4]:[0x3] = obj
	movff	ACCUL, TMP_REG_3 ; TMP_REG_4:3 <- count * 2 + 1
	movff	ACCUH, TMP_REG_4
	bcf	STATUS, C	 ; TMP_REG_6:5 <- limit + 1
	rrcf	[0x2], W
	movwf	TMP_REG_6
	rrcf	[0x1], W
	movwf	TMP_REG_5
	infsnz	TMP_REG_5, F
	incf	TMP_REG_6, F
	movff	[0x4], FSR0H	 ; FSR0 <- obj
	movff	[0x3], FSR0L
	movff	FSR2H, TMP_REG_2 ; TMP_REG_2:1 <- FSR2
	movff	FSR2L, TMP_REG_1
	clrf	ACCUH		 ; ACCU <- 0
	clrf	ACCUL
caml_hash_univ_param_beg:
	btfss	FSR0L, 0	 ; isint(ACCU) ?
	bra	caml_hash_univ_param_block ; no -> goto block
	movf	FSR0L, W	 ; yes -> hash(int)
	rcall	caml_hash_univ_param_routine
	movf	FSR0H, W
	rcall	caml_hash_univ_param_routine
caml_hash_univ_param_back:
	movf	FSR2L, W	 ; check stack top
	cpfseq	TMP_REG_1
	bra	caml_hash_univ_param_pop
	movf	FSR2H, W
	cpfslt	TMP_REG_2
	bra	caml_hash_univ_param_return
caml_hash_univ_param_pop:	 ; go backward
	movff	PREINC2, TMP_REG_7
	movff	PREINC2, FSR0L
	movff	PREINC2, FSR0H
	bra	caml_hash_univ_param_loop
caml_hash_univ_param_block:
	subfsr	FSR0, 0x2	 ; FSR0 <- @tag
	movf	POSTINC0, W 	 ; W <- tag ; FSR0 ++
	xorlw	STRING_TAG	 ; string ?
	bz	caml_hash_univ_param_data
	xorlw	(DOUBLE_TAG ^ STRING_TAG) 	  ; float ?
	bz	caml_hash_univ_param_data
	xorlw	(DOUBLE_ARRAY_TAG ^ DOUBLE_TAG)   ; float[] ?
	bz	caml_hash_univ_param_data
	xorlw	(ABSTRACT_TAG ^ DOUBLE_ARRAY_TAG) ; abstract ?
	bz	caml_hash_univ_param_back
	xorlw	(INFIX_TAG ^ ABSTRACT_TAG) 	  ; infix ?
	bz	caml_hash_univ_param_infix
	xorlw	(FORWARD_TAG ^ INFIX_TAG) 	  ; forward ?
	bz	caml_hash_univ_param_forward
	xorlw	(OBJECT_TAG ^ FORWARD_TAG) 	  ; object ?
	bz	caml_hash_univ_param_object
	xorlw	(CUSTOM_TAG ^ OBJECT_TAG) 	  ; custom ?
	bz	caml_hash_univ_param_custom
caml_hash_univ_param_beg_loop:
	rcall	caml_hash_univ_param_routine 	; hash(tag)
	movff	POSTINC0, TMP_REG_7		; TMP_REG_7 <- size
caml_hash_univ_param_loop:
	decf	TMP_REG_5, F			; check limit
	bnz	caml_hash_univ_param_loop_continue
	decf	TMP_REG_6, F
	bz	caml_hash_univ_param_popstop
caml_hash_univ_param_loop_continue:
	movf	TMP_REG_7, W			; last field ?
	bz	caml_hash_univ_param_back	; yes -> go back
	movf	POSTINC0, W			; isint(field) ?
	btfss	WREG, 0
	bra	caml_hash_univ_param_loop_block ; no -> goto loop_block
	rcall	caml_hash_univ_param_routine	; yes -> hash(int)
	movf	POSTINC0, W
	rcall	caml_hash_univ_param_routine
	decf	TMP_REG_7, F			; size --;
	bra	caml_hash_univ_param_loop	; loop
caml_hash_univ_param_loop_block:
	movwf	TMP_REG_8			; TMP_REG_8 <- fieldL
	dcfsnz	TMP_REG_7, F			; size --; if size = 0
	bra	caml_hash_univ_param_loop_block_last ; then goto loop_block_last
	subfsr	FSR2, 3				; push3
	movlw	(STACK_END - 0x1)		; check STACK OVERFLOW
	cpfsgt	FSR2H
	bra	caml_raise_stack_overflow
	movf	TMP_REG_7, W			; push (size, FSR0)
	movwf	[0x1]
	movf	FSR0L, W
	movwf	[0x2]
	movf	FSR0H, W
	movwf	[0x3]
caml_hash_univ_param_loop_block_last:
	movf	INDF0, W			; forward
	movwf	FSR0H
	movff	TMP_REG_8, FSR0L
	bra	caml_hash_univ_param_beg
caml_hash_univ_param_data:			; hash all data block
	incf	POSTINC0, W			; TMP_REG_8 <- size + 1
	movwf	TMP_REG_8
	rcall	caml_hash_univ_param_routine 	; hash(size)
caml_hash_univ_param_data_loop:
	dcfsnz	TMP_REG_8, F			; if TMP_REG_8 -- = 0
	bra	caml_hash_univ_param_back	; then go back
	movf	POSTINC0, W			; hash fieldL
	rcall	caml_hash_univ_param_routine
	movf	POSTINC0, W			; hash fieldH
	rcall	caml_hash_univ_param_routine
	bra	caml_hash_univ_param_data_loop 	; loop
caml_hash_univ_param_infix:
	movf	INDF0, W	; FSR0 -= 2 * (*FSR0)
	subwf	FSR0L, F
	btfss	STATUS, C
	decf	FSR0H, F
	subwf	FSR0L, F
	btfss	STATUS, C
	decf	FSR0H, F
	bra	caml_hash_univ_param_beg_loop
caml_hash_univ_param_forward:
	addfsr	FSR0, 1		    ; FSR0++
	movff	POSTINC0, TMP_REG_8 ; FSR0 <- *FSR0
	movf	INDF0, W
	movwf	FSR0H
	movff	TMP_REG_8, FSR0L
	bra	caml_hash_univ_param_beg ; hash forwarded
caml_hash_univ_param_object:
	addfsr	FSR0, 3		; FSR0 += 3
	movf	POSTINC0, W	; hash(object_id)
	rcall	caml_hash_univ_param_routine
	movf	INDF0, W
	rcall	caml_hash_univ_param_routine
	bra	caml_hash_univ_param_back ; go back
caml_hash_univ_param_custom:
	bra	caml_hash_univ_param_back
caml_hash_univ_param_popstop:
	pop
caml_hash_univ_param_stop:
	movff	TMP_REG_1, FSR2L
	movff	TMP_REG_2, FSR2H
caml_hash_univ_param_return:
	bsf	ACCUL, 0
	return
caml_hash_univ_param_routine:
	decf	TMP_REG_3, F
	bnz	caml_hash_univ_param_routine_continue
	decf	TMP_REG_4, F
	bz	caml_hash_univ_param_popstop
caml_hash_univ_param_routine_continue:
	addwf	ACCUL, F
	btfsc	STATUS, C
	incf	ACCUH, F
	movlw	0xDF
	mulwf	ACCUL
	movff	PRODL, ACCUL
	movf	PRODH, W
	addwf	ACCUH, F
	return
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;               OBJ                 ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_obj_dup
caml_obj_dup:
	;; ACCU = obj
	movff	ACCUL, FSR0L
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1
	movf	POSTDEC0, W
	bz	caml_obj_dup_end
	movwf	TABLAT
	call	_CHECK_UNFULL_HEAP_GEN
	movff	ACCUL, FSR0L
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x2
	movff	POSTINC0, POSTINC1 ; copy tag
	movff	POSTINC0, POSTINC1 ; copy size
	movff	FSR1L, ACCUL
	movff	FSR1H, ACCUH
caml_obj_dup_loop:
	movff	POSTINC0, POSTINC1 ; copy 1 element
	movff	POSTINC0, POSTINC1
	decfsz	TABLAT, F	   ; loop size times
	bra	caml_obj_dup_loop
caml_obj_dup_end:
	return
#endif

#ifdef caml_useprim_caml_obj_block
caml_obj_block:
	;; ACCU = tag
	;; [0x2]:[0x1] = size
	rrcf	[0x2], W	; check allocation
	rrcf	[0x1], W
	bz	caml_obj_block_atom
	movwf	TABLAT
	call	_CHECK_UNFULL_HEAP_GEN
	rrcf	ACCUH, W	; write tag
	rrcf	ACCUL, W
	movwf	POSTINC1
	rrcf	[0x2], W	; write size
	rrcf	[0x1], W
	movwf	POSTINC1
	movff	FSR1L, ACCUL
	movff	FSR1H, ACCUH
caml_obj_block_loop:		; fill with unit
	clrf	INDF1
	bsf	POSTINC1, 0
	clrf	POSTINC1
	decfsz	WREG, F		; loop size times
	bra	caml_obj_block_loop
	return
caml_obj_block_atom:
	movlw	low (ATOM0_SIZE + 1)
	movwf	ACCUL
	movlw	high (ATOM0_SIZE + 1)
	movwf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_obj_set_tag
caml_obj_set_tag:
	;; ACCU = obj
	;; [0x2]:[0x1] = tag
	movff	ACCUL, FSR0L	; FSR0 <- obj
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x2	; FSR0 <- @old_tag
	rrcf	[0x2], W	; W <- new_tag
	rrcf	[0x1], W
	movwf	INDF0		; write new_tag
	return
#endif

#ifdef caml_useprim_caml_obj_tag
caml_obj_tag:
	btfsc	ACCUL, 0		; is int ?
	bra	caml_obj_tag_int	; yes -> return Int
	movf	ACCUH, W		; in heap ?
	subwf	CUR_HEAP_END, W
	bnc	caml_obj_tag_out_of_memory ; no -> return Out_of_heap
	addlw	-HEAP_SIZE
	bc	caml_obj_tag_out_of_memory ; no -> return Out_of_heap
	movff	ACCUL, FSR0L		; FSR0 <- obj
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x2		; FSR0 <- @tag
	clrf	ACCUH			; ACCU <- tag
	bsf	STATUS, C
	rlcf	INDF0, W
	rlcf	ACCUH, F
	movwf	ACCUL
	return
caml_obj_tag_int:
	movlw	low ((OBJ_TAG_INT * 0x2) + 0x1)
	movwf	ACCUL
	movlw	high ((OBJ_TAG_INT * 0x2) + 0x1)
	movwf	ACCUH
	return
caml_obj_tag_out_of_memory:
	movlw	low ((OBJ_TAG_OUT_OF_MEMORY * 0x2) + 0x1)
	movwf	ACCUL
	movlw	high ((OBJ_TAG_OUT_OF_MEMORY * 0x2) + 0x1)
	movwf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_obj_is_block
caml_obj_is_block:
	movlw	0x1
	btfss	ACCUL, 0
	movlw	0x3
	movwf	ACCUL
	clrf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_obj_truncate
#ifndef caml_useprim_caml_raise_ia_obj_truncate
#define caml_useprim_caml_raise_ia_obj_truncate
#endif
caml_obj_truncate:
	;; ACCU = obj
	;; [0x2]:[0x1] = size
	movff	ACCUL, FSR0L	; FSR0 <- obj
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1	; FSR0 <- @old_size
	bcf	STATUS, C	; W <- new_size
	rrcf	[0x2], W
	bnz	caml_raise_ia_obj_truncate ; if new_size > 255 -> raise
	rrcf	[0x1], W
	bz	caml_raise_ia_obj_truncate ; if new_size = 0 -> raise
	subwf	INDF0, W	; W <- old_size - new_size
	bnc	caml_raise_ia_obj_truncate ; if new_s > old_s -> raise
	subwf	INDF0, F	; write (old_size - (old_size - new_size))
	return
#endif

#ifdef caml_useprim_caml_raise_ia_obj_truncate
#ifndef caml_useprim_caml_raise_ia
#define caml_useprim_caml_raise_ia
#endif
caml_raise_ia_obj_truncate:
	M_CHECK_UNFULL_HEAP 0xA		; check allocation of string and exn
	M_WRITE_BYTE STRING_TAG		; write string tag
	M_WRITE_BYTE 0x7		; write string size
	movff	FSR1L, TMP_REG_1	; mem string addr
	movff	FSR1H, TMP_REG_2
	M_WRITE_BYTE 'O'		; write string content
	M_WRITE_BYTE 'b'
	M_WRITE_BYTE 'j'
	M_WRITE_BYTE '.'
	M_WRITE_BYTE 't'
	M_WRITE_BYTE 'r'
	M_WRITE_BYTE 'u'
	M_WRITE_BYTE 'n'
	M_WRITE_BYTE 'c'
	M_WRITE_BYTE 'a'
	M_WRITE_BYTE 't'
	M_WRITE_BYTE 'e'
	M_WRITE_BYTE 0x0		; write string \0\1
	M_WRITE_BYTE 0x1
	goto	caml_raise_ia
#endif


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;                GC                 ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_heap_size
caml_heap_size:
	movlw	low (((HEAP_SIZE * 0x100) * 0x2) + 0x1)
	movwf	ACCUL
	movlw	high (((HEAP_SIZE * 0x100) * 0x2) + 0x1)
	movwf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_stack_size
caml_stack_size:
	movlw	low (((STACK_ANCHOR - 0xEFF) + (STACK_SIZE * 0x100)) + 0x1)
	movwf	ACCUL
	movlw	high (((STACK_ANCHOR - 0xEFF) + (STACK_SIZE * 0x100)) + 0x1)
	movwf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_heap_occupation
caml_heap_occupation:
	movf	CUR_HEAP_ANCHOR, W
	subwf	FSR1H, W
	movwf	ACCUH
	bsf	STATUS, C
	rlcf	FSR1L, W
	movwf	ACCUL
	rlcf	ACCUH, F
	return
#endif

#ifdef caml_useprim_caml_stack_occupation
caml_stack_occupation:
	movf	FSR2L, W
	sublw	low (STACK_ANCHOR + 0x1)
	movwf	ACCUL
	movlw	high (STACK_ANCHOR + 0x1)
	subfwb	FSR2H, W
	movwf	ACCUH
	bsf	ACCUL, 0
	return
#endif

#ifdef caml_useprim_caml_running_number
caml_running_number:
	bsf	STATUS, C
	rlcf	GC_COUNTER_L, W
	movwf	ACCUL
	rlcf	GC_COUNTER_H, W
	movwf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_gc_run
caml_gc_run:
	goto	caml_gc_exec
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;              RANDOM               ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_random_bits
#ifndef caml_useprim_caml_random_round
#define caml_useprim_caml_random_round
#endif
caml_random_bits:
	rcall	caml_random_round
	movff	RAND_CUR_L, ACCUL
	movff	RAND_CUR_H, ACCUH
	bsf	ACCUL, 0
	bcf	ACCUH, 7
	return
#endif

#ifdef caml_useprim_caml_random_bool
#ifndef caml_useprim_caml_random_round
#define caml_useprim_caml_random_round
#endif
caml_random_bool:
	rcall	caml_random_round
	clrf	ACCUH
	movf	RAND_CUR_L, W
	andlw	0x03
	iorlw	0x01
	movwf	ACCUL
	return
#endif

#ifdef caml_useprim_caml_random_round
caml_random_round:
	movlw	0x4
	movwf	TMP_REG_1
caml_random_round_loop:
	movff	RAND_CUR_L, FSR0L
	movff	RAND_CUR_H, FSR0H
	movf	POSTINC0, W
	xorlw	0xAA
	xorwf	RAND_CUR_L, F
	movf	POSTINC0, W
	xorlw	0xAA
	xorwf	RAND_CUR_H, F
	swapf	RAND_CUR_L, F
	swapf	RAND_CUR_H, F
	rlcf	RAND_CUR_L, W
	rlcf	RAND_CUR_H, F
	rlcf	RAND_CUR_L, F
	movlw	.223
	addwf	RAND_CUR_L, F
	addwf	RAND_CUR_H, F
	decfsz	TMP_REG_1, F
	bra	caml_random_round_loop
	return
#endif


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;            SFR ACCESS             ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_pic_write_reg
caml_pic_write_reg:
	bsf	STATUS, C	; FSR0 <- 0x0F70 | (ACCUL >> 1)
	rrcf	ACCUL, W
	movwf	FSR0L
	setf	FSR0H
	rrcf	[0x2], W	; STATUS.C ignored
	rrcf	[0x1], W	; INDF0 <- stack[0] >> 1
	movwf	INDF0
	M_CONST	0
	return
#endif

#ifdef caml_useprim_caml_pic_read_reg
caml_pic_read_reg:
	bsf	STATUS, C	; FSR0 <- 0x0F70 | (ACCUL >> 1)
	rrcf	ACCUL, W
	movwf	FSR0L
	setf	FSR0H
	rlcf	INDF0, W	; ACCU <- (INDF0 << 1) | 1
	movwf	ACCUL
	rlcf	ACCUH, F
	return
#endif

#ifdef caml_useprim_caml_pic_set_bit
caml_pic_set_bit:
	bsf	STATUS, C	; FSR0 <- 0xF70 | (ACCUL >> 1)
	rrcf	ACCUL, W
	movwf	FSR0L
	setf	FSR0H
	movf	INDF0, W	; INDF0 <- INDF0 | ACCUH
	iorwf	ACCUH, W
	movwf	INDF0
	M_CONST	0
	return
#endif

#ifdef caml_useprim_caml_pic_clear_bit
caml_pic_clear_bit:
	bsf	STATUS, C	; FSR0 <- 0xF70 | (ACCUL >> 1)
	rrcf	ACCUL, W
	movwf	FSR0L
	setf	FSR0H
	comf	ACCUH, F	; INDF0 <- INDF0 & ~ACCUH
	movf	INDF0, W
	andwf	ACCUH, W
	movwf	INDF0
	M_CONST	0
	return
#endif

#ifdef caml_useprim_caml_pic_test_bit
caml_pic_test_bit:
	bsf	STATUS, C	; FSR0 <- 0xF70 | (ACCUL >> 1)
	rrcf	ACCUL, W
	movwf	FSR0L
	setf	FSR0H
	movf	ACCUH, W	; INDF0 & ACCUH
	andwf	INDF0, W
	movlw	0x1		; ACCU <- if Z then 1 else 3
	btfss	STATUS, Z
	movlw	0x3
	movwf	ACCUL
	clrf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_pic_tris_of_port
caml_pic_tris_of_port:
	movlw	((0x2 * LATA) & 0xFF)
	cpfslt	ACCUL
	addlw	(0x2 * (PORTA - LATA))
	addlw	((0x2 * (TRISA - PORTA) - (0x2 * LATA)) & 0xFF)
	addwf	ACCUL, F
	return
#endif


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;             FORMATS               ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_int_of_string
#ifndef caml_useprim_caml_raise_ios_failure
#define caml_useprim_caml_raise_ios_failure
#endif
caml_int_of_string:
	movff	ACCUL, FSR0L	; FSR0 <- ACCU
	movff	ACCUH, FSR0H
	clrf	ACCUL		; ACCU <- 0
	clrf	ACCUH
	movf	POSTINC0, W
	bz	caml_raise_ios_failure
	sublw	'-'		; W <- '-' - str[0]
	bcf	TMP_REG_5, 0	; TMP_REG_5.0 = 0  <=>  res >= 0
	bnz	caml_int_of_string_sign_tested
	bsf	TMP_REG_5, 0
	movf	POSTINC0, W
caml_int_of_string_sign_tested:
	sublw	('-' - '0')	; W <- str[0] - '0'
	bnz	caml_int_of_string_dec
	movf	POSTINC0, W	; W <- str[1]
	bz	caml_int_of_string_return_zero
	sublw	'x'		; W <- 'x' - str[1]
	bz	caml_int_of_string_hex
	sublw	('x' - 'X')	; W <- str[1] - 'X'
	bz	caml_int_of_string_hex
	sublw	('b' - 'X')	; W <- 'b' - str[1]
	bz	caml_int_of_string_bin
	sublw	('b' - 'B')	; W <- str[1] - 'B'
	bz	caml_int_of_string_bin
	sublw	('o' - 'B')	; W <- 'o' - str[1]
	bz	caml_int_of_string_oct
	sublw	('o' - 'O')	; W <- str[1] - 'O'
	bz	caml_int_of_string_oct
	addlw	('O' - '0')	; W <- str[1] - '0'
caml_int_of_string_dec:
	sublw	('9' - '0')	; W <- ('9' - '0') - (c - '0')
	bnc	caml_raise_ios_failure
	sublw	('9' - '0')	; W <- ('9' - '0') - ('9' - c)
	bnc	caml_raise_ios_failure
	movwf	TMP_REG_1
	movf	ACCUL, W
	mullw	.10
	movf	TMP_REG_1, W
	addwf	PRODL, W
	movwf	ACCUL
	movf	PRODH, W
	addwfc	ACCUH, F
	bn	caml_raise_ios_failure
	movf	POSTINC0, W	; W <- str[i++]
	bz	caml_int_of_string_return
	addlw	(-'0')		; W <- c - '0'
	bra	caml_int_of_string_dec
caml_int_of_string_oct:
	movf	POSTINC0, W	; W <- str[i++]
	bz	caml_raise_ios_failure
caml_int_of_string_oct_loop:
	addlw	(-'0')		; W <- c - '0'
	swapf	WREG, F
	rlncf	WREG, F
	rlcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	rlcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	rlcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	andlw	B'11111000'	; check range
	bnz	caml_raise_ios_failure
	movf	POSTINC0, W	; W <- str[i++]
	bnz	caml_int_of_string_oct_loop
	bra	caml_int_of_string_return
caml_int_of_string_bin:
	movf	POSTINC0, W	; W <- str[i++]
	bz	caml_raise_ios_failure
caml_int_of_string_bin_loop:
	addlw	(-'0')		; W <- c - '0'
	rrcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	andlw	B'01111111'	; check range
	bnz	caml_raise_ios_failure
	movf	POSTINC0, W	; W <- str[i++]
	bnz	caml_int_of_string_bin_loop
	bra	caml_int_of_string_return
caml_int_of_string_hex:
	movf	POSTINC0, W	; W <- str[i++]
	bz	caml_raise_ios_failure
caml_int_of_string_hex_loop:
	addlw	(-'0')		; W <- c - '0'
	bnc	caml_raise_ios_failure
	addlw	('0' - '9' - 1)	; W <- (c - '0') + ('0' - '9' - 1)
	bnc	caml_int_of_string_hex_09
	addlw	('9' - 'A' + 1)	; W <- (c - '9' - 1) + ('9' - 'A' + 1)
	bnc	caml_raise_ios_failure
	addlw	('A' - 'F' - 1)	; W <- (c - 'A') + ('A' - 'F' - 1)
	bnc	caml_int_of_string_hex_AF
	addlw	('F' - 'a' + 1)	; W <- (c - 'F' - 1) + ('F' - 'a' + 1)
	bnc	caml_raise_ios_failure
	addlw	('a' - 'f' - 1)	; W <- (c - 'a') + ('a' - 'f' - 1)
	bc	caml_raise_ios_failure
caml_int_of_string_hex_af:
	addlw	('f' - 'F' + 'A' - 'a')
caml_int_of_string_hex_AF:
	addlw	('F' - '9' + '0' - 'A' + .10)
caml_int_of_string_hex_09:
	addlw	('9' - '0' + 1)
caml_int_of_string_hex_shift:
	swapf	WREG, F
	rlcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	rlcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	rlcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	rlcf	WREG, F
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	bn	caml_raise_ios_failure
	movf	POSTINC0, W
	bnz	caml_int_of_string_hex_loop
caml_int_of_string_return:
	bsf	STATUS, C
	rlcf	ACCUL, F
	rlcf	ACCUH, F
	btfss	TMP_REG_5, 0	; negate ?
	return			; no -> return
	comf	ACCUL, F	; yes -> ACCU <- -ACCU
	comf	ACCUH, F
	infsnz	ACCUL, F
	incf	ACCUH, F
	return
caml_int_of_string_return_zero:
	bsf	ACCUL, 0
	return
#endif

#ifdef caml_useprim_caml_raise_ios_failure
caml_raise_ios_failure:
	movlw	0x1			; erase ACCU
	movwf	ACCUL
	clrf	ACCUH
	M_CHECK_UNFULL_HEAP 0xA		; check allocation of string and exn
	M_WRITE_BYTE STRING_TAG		; write string tag
	M_WRITE_BYTE 0x7		; write string size
	movff	FSR1L, TMP_REG_1	; mem string addr
	movff	FSR1H, TMP_REG_2
	M_WRITE_BYTE 'i'		; write string content
	M_WRITE_BYTE 'n'
	M_WRITE_BYTE 't'
	M_WRITE_BYTE '_'
	M_WRITE_BYTE 'o'
	M_WRITE_BYTE 'f'
	M_WRITE_BYTE '_'
	M_WRITE_BYTE 's'
	M_WRITE_BYTE 't'
	M_WRITE_BYTE 'r'
	M_WRITE_BYTE 'i'
	M_WRITE_BYTE 'n'
	M_WRITE_BYTE 'g'
	M_WRITE_BYTE 0x0		; write string \0
	M_WRITE_BYTE 0x0		; write exn tag
	M_WRITE_BYTE 0x2		; write exn size
	movff	FSR1L, ACCUL		; ACCU <- exn
	movff	FSR1H, ACCUH
	lfsr	FSR0, (STACK_ANCHOR - 0x1 - (0x2 * FAILURE_IND))
	movff	POSTINC0, POSTINC1 	; write global(FAILURE_IND)
	movff	POSTINC0, POSTINC1
	movff	TMP_REG_1, POSTINC1 	; write string addr
	movff	TMP_REG_2, POSTINC1
	goto	_RAISE
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;              FLOAT                ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_neg_float
caml_neg_float:
	M_CHECK_UNFULL_HEAP 0x1
	movff	ACCUL, FSR0L		; FSR0 <- @f
	movff	ACCUH, FSR0H
	movlw	DOUBLE_TAG		; write tag
	movwf	POSTINC1
	movlw	0x1			; write size
	movwf	POSTINC1
	movff	FSR1L, ACCUL		; ACCU <- @result
	movff	FSR1H, ACCUH
	movff	POSTINC0, POSTINC1 	; copy float
	movff	INDF0, INDF1
	btg	INDF1, 7		; inverse sign bit
	return
#endif

#ifdef caml_useprim_caml_abs_float
caml_abs_float:
	M_CHECK_UNFULL_HEAP 0x1
	movff	ACCUL, FSR0L		; FSR0 <- @f
	movff	ACCUH, FSR0H
	movlw	DOUBLE_TAG		; write tag
	movwf	POSTINC1
	movlw	0x1			; write size
	movwf	POSTINC1
	movff	FSR1L, ACCUL		; ACCU <- @result
	movff	FSR1H, ACCUH
	movff	POSTINC0, POSTINC1 	; copy float
	movff	INDF0, INDF1
	bcf	INDF1, 7		; clear sign bit
	return
#endif

#ifdef caml_useprim_caml_add_float
#ifndef caml_useprim_caml_prepare_floats_calculus
#define caml_useprim_caml_prepare_floats_calculus
#endif
caml_add_float:
	rcall	caml_prepare_floats_calculus
	movf	TMP_REG_3, W
	subwf	TMP_REG_7, W		; e2 - e1
	bz	caml_add_float_meq
	bnc	caml_add_float_loop_e1lte2
caml_add_float_loop_e1gte2
	bcf	STATUS, C		; m2 >>= 1
	rrcf	TMP_REG_6, F
	rrcf	TMP_REG_5, F
	incf	TMP_REG_7, F		; incr e2
	movf	TMP_REG_7, W		; while e2 != e1
	cpfseq	TMP_REG_3
	bra	caml_add_float_loop_e1gte2
	bra	caml_add_float_meq
caml_add_float_loop_e1lte2:
	bcf	STATUS, C		; m1 >>= 1
	rrcf	TMP_REG_2, F
	rrcf	TMP_REG_1, F
	incf	TMP_REG_3, F		; incr e1
	movf	TMP_REG_3, W		; while e1 != e2
	cpfseq	TMP_REG_7
	bra	caml_add_float_loop_e1lte2
caml_add_float_meq:
	btfsc	TMP_REG_4, 7		; f1 negative ?
	bra	caml_add_float_NX
	btfsc	TMP_REG_8, 7
	bra	caml_add_float_PN
caml_add_float_PP:
	movf	TMP_REG_1, W
	addwf	TMP_REG_5, W		; m1L + m2L
	movwf	POSTINC1		; write mRL
	movf	TMP_REG_2, W
	addwfc	TMP_REG_6, W		; m1H + m2H
	btfss	WREG, 3			; carry ?
	bra	caml_add_float_PP_no_carry
	rrcf	WREG, F			; shift mR
	subfsr	FSR1, 0x1
	rrcf	POSTINC1, F
	incf	TMP_REG_3, F		; incr eR
caml_add_float_PP_no_carry:
	movwf	INDF1			; write mRH
	swapf	INDF1, F
	rlcf	INDF1, F
	movf	TMP_REG_3, W
	addlw	(-0x80)
	iorwf	INDF1, F		; write eR
	bcf	INDF1, 7		; write sR
	return
caml_add_float_NN:
	movf	TMP_REG_1, W
	addwf	TMP_REG_5, W		; m1L + m2L
	movwf	POSTINC1		; write mRL
	movf	TMP_REG_2, W
	addwfc	TMP_REG_6, W		; m1H + m2H
	btfss	WREG, 3			; carry ?
	bra	caml_add_float_NN_no_carry
	rrcf	WREG, F			; shift mR
	subfsr	FSR1, 0x1
	rrcf	POSTINC1, F
	incf	TMP_REG_3, F		; incr eR
caml_add_float_NN_no_carry:
	movwf	INDF1			; write mRH
	swapf	INDF1, F
	rlcf	INDF1, F
	movf	TMP_REG_3, W
	addlw	(-0x80)
	iorwf	INDF1, F		; write eR
	bsf	INDF1, 7		; write sR
	return
caml_add_float_NX:
	btfsc	TMP_REG_8, 7
	bra	caml_add_float_NN
caml_add_float_NP:
	movf	TMP_REG_5, W		; m1 <-> m2
	movff	TMP_REG_1, TMP_REG_5
	movwf	TMP_REG_1
	movf	TMP_REG_6, W
	movff	TMP_REG_2, TMP_REG_6
	movwf	TMP_REG_2
caml_add_float_PN:
	movf	TMP_REG_5, W
	subwf	TMP_REG_1, W		; m1L - m2L
	movwf	TMP_REG_9		; TMP_REG_9 <- mR
	movf	TMP_REG_6, W
	subwfb	TMP_REG_2, W		; m1H - m2H
	movwf	TMP_REG_A
	iorwf	TMP_REG_9, W		; 0 ?
	bz	caml_add_float_return_zero
	bcf	TMP_REG_4, 7		; TMP_REG_4.7 <- result sign
	btfss	TMP_REG_A, 7		; < 0 ?
	bra	caml_add_float_not_neg
	comf	TMP_REG_9, F		; negate TMP_REG_A:9
	comf	TMP_REG_A, F
	infsnz	TMP_REG_9, F
	incf	TMP_REG_A, F
	bsf	TMP_REG_4, 7		; set result sign
caml_add_float_not_neg:
	btfsc	TMP_REG_A, 2
	bra	caml_add_float_end_normalize
caml_add_float_loop_normalize:
	bcf	STATUS, C
	rlcf	TMP_REG_9, F
	rlcf	TMP_REG_A, F
	incf	TMP_REG_3, F
	btfss	TMP_REG_A, 2
	bra	caml_add_float_loop_normalize
caml_add_float_end_normalize:
	movff	TMP_REG_9, POSTINC1 	; write low mantissa
	movf	TMP_REG_A, W		; write high mantissa
	andlw	B'00000011'
	movwf	INDF1
	swapf	INDF1, F
	bcf	STATUS, C
	rlcf	INDF1, F
	btfsc	TMP_REG_4, 7		; adjust result sign
	bsf	INDF1, 7
	movf	TMP_REG_3, W		; write result exponent
	addlw	(-0x80)
	iorwf	INDF1, F
	return
caml_add_float_return_zero:
	clrf	POSTINC1
	clrf	POSTINC1
	return
#endif

#ifdef caml_useprim_caml_prepare_floats_calculus
caml_prepare_floats_calculus:
	;; TMP_REG_1 = m1L
	;; TMP_REG_2 = m1H
	;; TMP_REG_3 = e1 + 0x80
	;; TMP_REG_4.7 = s1
	;; TMP_REG_5 = m2L
	;; TMP_REG_6 = m2H
	;; TMP_REG_7 = e2 + 0x80
	;; TMP_REG_8.7 = s2
	M_CHECK_UNFULL_HEAP 0x1
	movff	ACCUL, FSR0L		; FSR0 <- @f1
	movff	ACCUH, FSR0H
	movlw	DOUBLE_TAG		; write tag
	movwf	POSTINC1
	movlw	0x1			; write size
	movwf	POSTINC1
	movff	FSR1L, ACCUL		; ACCU <- @result
	movff	FSR1H, ACCUH
	movff	POSTINC0, TMP_REG_1 	; TMP_REG_1 <- m1L
	movf	INDF0, W		; TMP_REG_4.7 <- s1
	movwf	TMP_REG_4
	andlw	B'00011111'		; TMP_REG_3 <- e1
	btfsc	INDF0, 4
	iorlw	B'11100000'
	addlw	0x80
	movwf	TMP_REG_3
	clrf	TMP_REG_2		; TMP_REG_2 <- m1H
	btfsc	INDF0, 5
	bsf	TMP_REG_2, 0
	btfsc	INDF0, 6
	bsf	TMP_REG_2, 1
	bsf	TMP_REG_2, 2
	movsf	[0x1], FSR0L		; FSR0 <- @f2
	movsf	[0x2], FSR0H
	movff	POSTINC0, TMP_REG_5 	; TMP_REG_5 <- m2L
	movf	INDF0, W		; TMP_REG_8.7 <- s2
	movwf	TMP_REG_8
	andlw	B'00011111'		; TMP_REG_7 <- e2
	btfsc	INDF0, 4
	iorlw	B'11100000'
	addlw	0x80
	movwf	TMP_REG_7
	clrf	TMP_REG_6		; TMP_REG_6 <- m2H
	btfsc	INDF0, 5
	bsf	TMP_REG_6, 0
	btfsc	INDF0, 6
	bsf	TMP_REG_6, 1
	bsf	TMP_REG_6, 2
	return
#endif


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;             COMPARE               ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_equal
#ifndef caml_useprim_caml_compare
#define caml_useprim_caml_compare
#endif
#ifndef caml_useprim_caml_compare_to_eq
#define caml_useprim_caml_compare_to_eq
#endif
caml_equal:
	rcall	caml_compare
	bra	caml_compare_to_eq
#endif

#ifdef caml_useprim_caml_eq_float
#ifndef caml_useprim_caml_float_compare
#define caml_useprim_caml_float_compare
#endif
#ifndef caml_useprim_caml_compare_to_eq
#define caml_useprim_caml_compare_to_eq
#endif
caml_eq_float:
	rcall	caml_float_compare
	bra	caml_compare_to_eq
#endif

#ifdef caml_useprim_caml_notequal
#ifndef caml_useprim_caml_compare
#define caml_useprim_caml_compare
#endif
#ifndef caml_useprim_caml_compare_to_neq
#define caml_useprim_caml_compare_to_neq
#endif
caml_notequal:
	rcall	caml_compare
	bra	caml_compare_to_neq
#endif

#ifdef caml_useprim_caml_neq_float
#ifndef caml_useprim_caml_float_compare
#define caml_useprim_caml_float_compare
#endif
#ifndef caml_useprim_caml_compare_to_neq
#define caml_useprim_caml_compare_to_neq
#endif
caml_neq_float:
	rcall	caml_float_compare
	bra	caml_compare_to_neq
#endif

#ifdef caml_useprim_caml_lessthan
#ifndef caml_useprim_caml_compare
#define caml_useprim_caml_compare
#endif
#ifndef caml_useprim_caml_compare_to_lt
#define caml_useprim_caml_compare_to_lt
#endif
caml_lessthan:
	rcall	caml_compare
	bra	caml_compare_to_lt
#endif

#ifdef caml_useprim_caml_string_lessthan
#ifndef caml_useprim_caml_string_compare
#define caml_useprim_caml_string_compare
#endif
#ifndef caml_useprim_caml_compare_to_lt
#define caml_useprim_caml_compare_to_lt
#endif
caml_string_lessthan:
	rcall	caml_string_compare
	bra	caml_compare_to_lt
#endif

#ifdef caml_useprim_caml_lt_float
#ifndef caml_useprim_caml_float_compare
#define caml_useprim_caml_float_compare
#endif
#ifndef caml_useprim_caml_compare_to_lt
#define caml_useprim_caml_compare_to_lt
#endif
caml_lt_float:
	rcall	caml_float_compare
	bra	caml_compare_to_lt
#endif

#ifdef caml_useprim_caml_lessequal
#ifndef caml_useprim_caml_compare
#define caml_useprim_caml_compare
#endif
#ifndef caml_useprim_caml_compare_to_le
#define caml_useprim_caml_compare_to_le
#endif
caml_lessequal:
	rcall	caml_compare
	bra	caml_compare_to_le
#endif

#ifdef caml_useprim_caml_string_lessequal
#ifndef caml_useprim_caml_string_compare
#define caml_useprim_caml_string_compare
#endif
#ifndef caml_useprim_caml_compare_to_le
#define caml_useprim_caml_compare_to_le
#endif
caml_string_lessequal:
	rcall	caml_string_compare
	bra	caml_compare_to_le
#endif

#ifdef caml_useprim_caml_le_float
#ifndef caml_useprim_caml_float_compare
#define caml_useprim_caml_float_compare
#endif
#ifndef caml_useprim_caml_compare_to_le
#define caml_useprim_caml_compare_to_le
#endif
caml_le_float:
	rcall	caml_float_compare
	bra	caml_compare_to_le
#endif

#ifdef caml_useprim_caml_greaterthan
#ifndef caml_useprim_caml_compare
#define caml_useprim_caml_compare
#endif
#ifndef caml_useprim_caml_compare_to_gt
#define caml_useprim_caml_compare_to_gt
#endif
caml_greaterthan:
	rcall	caml_compare
	bra	caml_compare_to_gt
#endif

#ifdef caml_useprim_caml_string_greaterthan
#ifndef caml_useprim_caml_string_compare
#define caml_useprim_caml_string_compare
#endif
#ifndef caml_useprim_caml_compare_to_gt
#define caml_useprim_caml_compare_to_gt
#endif
caml_string_greaterthan:
	rcall	caml_string_compare
	bra	caml_compare_to_gt
#endif

#ifdef caml_useprim_caml_gt_float
#ifndef caml_useprim_caml_float_compare
#define caml_useprim_caml_float_compare
#endif
#ifndef caml_useprim_caml_compare_to_gt
#define caml_useprim_caml_compare_to_gt
#endif
caml_gt_float:
	rcall	caml_float_compare
	bra	caml_compare_to_gt
#endif

#ifdef caml_useprim_caml_greaterequal
#ifndef caml_useprim_caml_compare
#define caml_useprim_caml_compare
#endif
#ifndef caml_useprim_caml_compare_to_ge
#define caml_useprim_caml_compare_to_ge
#endif
caml_greaterequal:
	rcall	caml_compare
	bra	caml_compare_to_ge
#endif

#ifdef caml_useprim_caml_string_greaterequal
#ifndef caml_useprim_caml_string_compare
#define caml_useprim_caml_string_compare
#endif
#ifndef caml_useprim_caml_compare_to_ge
#define caml_useprim_caml_compare_to_ge
#endif
caml_string_greaterequal:
	rcall	caml_string_compare
	bra	caml_compare_to_ge
#endif

#ifdef caml_useprim_caml_ge_float
#ifndef caml_useprim_caml_float_compare
#define caml_useprim_caml_float_compare
#endif
#ifndef caml_useprim_caml_compare_to_ge
#define caml_useprim_caml_compare_to_ge
#endif
caml_ge_float:
	rcall	caml_float_compare
	bra	caml_compare_to_ge
#endif

;;;

#ifdef caml_useprim_caml_compare
#ifndef caml_useprim_caml_int_compare
#define caml_useprim_caml_int_compare
#endif
#ifndef caml_useprim_caml_string_compare
#define caml_useprim_caml_string_compare
#endif
#ifndef caml_useprim_caml_float_compare
#define caml_useprim_caml_float_compare
#endif
#ifndef caml_useprim_caml_float_array_compare
#define caml_useprim_caml_float_array_compare
#endif
#ifndef caml_useprim_caml_raise_ia_compare
#define caml_useprim_caml_raise_ia_compare
#endif
caml_compare:
	;; ACCU = o1
	;; [0x2]:[0x1] = o2
	
;;; Fast traitment in simple cases
	movf	ACCUL, W		; == ?
	xorwf	[0x1], W
	bnz	caml_compare_A_phydiff
	movf	ACCUH, W
	xorwf	[0x2], W
	bz	caml_compare_A_1eq2
caml_compare_A_phydiff:
	btfss	ACCUL, 0		; is_int(o1) ?
	bra	caml_compare_A_addr_x	; no
	btfss	[0x1], 0	  	; yes ; is_int(o2) ?
	bra	caml_compare_A_1lt2	; no  ; compare int addr -> return -1
	bra	caml_int_compare 	; yes ; compare int int
caml_compare_A_addr_x:
	btfsc	[0x1], 0		; is_int(o2) ?
	bra	caml_compare_A_1gt2	; yes ; compare addr int -> return 1
	
	movff	FSR1L, TMP_REG_6	; save FSR1
	movff	FSR1H, TMP_REG_7
	movff	ACCUL, FSR1L		; FSR1 <- o1
	movff	ACCUH, FSR1H
	movsf	[0x1], FSR0L		; FSR0 <- o2
	movsf	[0x2], FSR0H
	subfsr	FSR1, 0x2		; FSR1 -= 2
	subfsr	FSR0, 0x2		; FSR0 -= 2
	
	movf	POSTINC1, W		; read tag(o1) ; FSR1 ++
	subwf	INDF0, W		; tag(o2) - tag(o1)
	bnc	caml_compare_B_1gt2
	bnz	caml_compare_B_1lt2

	movf	POSTINC0, W
	xorlw	STRING_TAG		; strings ?
	bz	caml_compare_B_string_compare
	xorlw	(STRING_TAG^DOUBLE_TAG) ; float ?
	bz	caml_compare_B_float_compare
	xorlw	(DOUBLE_TAG^DOUBLE_ARRAY_TAG) ; float[] ?
	bz	caml_compare_B_float_array_compare
	xorlw	(DOUBLE_ARRAY_TAG^ABSTRACT_TAG)	; abstract ?
	bz	caml_compare_B_raise_ia_compare
	xorlw	(ABSTRACT_TAG^CLOSURE_TAG) ; closure ?
	bz	caml_compare_B_raise_ia_compare
	xorlw	(CLOSURE_TAG^INFIX_TAG) ; infix ?
	bz	caml_compare_B_raise_ia_compare
	
	movf	POSTINC1, W		; size equals ? FSR1 ++
	movwf	TMP_REG_A		; TMP_REG_A <- size

	subwf	POSTINC0, W		; size(o2) - size(o1)
	bnc	caml_compare_B_1gt2
	bnz	caml_compare_B_1lt2

;;; Not a simple case -> go through data trees
	movff	FSR2L, TMP_REG_8 	; save FSR2
	movff	FSR2H, TMP_REG_9
	bra	caml_compare_loop

;;; Tools
caml_compare_raise_stack_overflow:
	movff	TMP_REG_6, FSR1L 	; restore FSR1
	movff	TMP_REG_7, FSR1H
	movff	TMP_REG_8, FSR2L 	; restore FSR2
	movff	TMP_REG_9, FSR2H
	goto	caml_raise_stack_overflow ; raise Stack_overflow
caml_compare_B_string_compare:
	movff	TMP_REG_6, FSR1L	; restore FSR1
	movff	TMP_REG_7, FSR1H
	bra	caml_string_compare 	; goto string_compare
caml_compare_B_float_compare:
	movff	TMP_REG_6, FSR1L	; restore FSR1
	movff	TMP_REG_7, FSR1H
	bra	caml_float_compare 	; goto float_compare
caml_compare_B_float_array_compare:
	movff	TMP_REG_6, FSR1L	; restore FSR1
	movff	TMP_REG_7, FSR1H
	bra	caml_float_array_compare; goto float_array_compare
caml_compare_CD_raise_ia_compare:
	movff	TMP_REG_8, FSR2L 	; restore FSR2
	movff	TMP_REG_9, FSR2H
caml_compare_B_raise_ia_compare:
	movff	TMP_REG_6, FSR1L 	; restore FSR1
	movff	TMP_REG_7, FSR1H
	goto	caml_raise_ia_compare 	; raise Invalid_argument "compare"
caml_compare_CD_1lt2:
	movff	TMP_REG_8, FSR2L 	; restore FSR2
	movff	TMP_REG_9, FSR2H
caml_compare_B_1lt2:
	movff	TMP_REG_6, FSR1L 	; restore FSR1
	movff	TMP_REG_7, FSR1H
caml_compare_A_1lt2:
	setf	ACCUL			; return -1
	setf	ACCUH
	return
caml_compare_CD_1gt2:
	movff	TMP_REG_8, FSR2L 	; restore FSR2
	movff	TMP_REG_9, FSR2H
caml_compare_B_1gt2:
	movff	TMP_REG_6, FSR1L 	; restore FSR1
	movff	TMP_REG_7, FSR1H
caml_compare_A_1gt2:
	M_CONST	1			; return 1
	return
caml_compare_CD_1ne2:
	movff	TMP_REG_8, FSR2L 	; restore FSR2
	movff	TMP_REG_9, FSR2H
	movff	TMP_REG_6, FSR1L 	; restore FSR1
	movff	TMP_REG_7, FSR1H
	return
caml_compare_CD_1eq2:
	movff	TMP_REG_6, FSR1L 	; restore FSR1
	movff	TMP_REG_7, FSR1H
caml_compare_A_1eq2:
	M_CONST	0			; return 0
	return
	
;;; Loop
caml_compare_forward:
	movff	POSTINC0, TMP_REG_A	; TMP_REG_A <- size ; FSR0 ++
caml_compare_loop:
	movf	POSTINC0, W		; stack[0] <- *FSR0++
	movwf	[0x1]
	movf	POSTINC0, W
	movwf	[0x2]
	movff	POSTINC1, ACCUL		; ACCU <- *FSR1++
	movff	POSTINC1, ACCUH

;;; Simple cases
	movf	ACCUL, W		; == ?
	xorwf	[0x1], W
	bnz	caml_compare_C_phydiff
	movf	ACCUH, W
	xorwf	[0x2], W
	bz	caml_compare_C_1eq2 	; if e1 == e2 -> branch
caml_compare_C_phydiff:
	btfss	ACCUL, 0		; is_int(e1) ?
	bra	caml_compare_C_addr_x	; no
	btfss	[0x1], 0	  	; yes ; is_int(e2) ?
	bra	caml_compare_CD_1lt2	; no  ; compare int addr -> return -1
	rcall	caml_int_compare 	; yes ; compare int int
	bra	caml_compare_C_compared
caml_compare_C_addr_x:
	btfsc	[0x1], 0		; is_int(e2) ?
	bra	caml_compare_CD_1gt2	; yes ; compare addr int -> return 1

	movff	FSR0L, TMP_REG_B 	; save FSR0 and FSR1
	movff	FSR0H, TMP_REG_C
	movff	FSR1L, TMP_REG_D
	movff	FSR1H, TMP_REG_E
	movsf	[0x1], FSR0L		; FSR0 <- stack[0]
	movsf	[0x2], FSR0H
	movff	ACCUL, FSR1L		; FSR1 <- ACCU
	movff	ACCUH, FSR1H
	subfsr	FSR1, 0x2		; FSR1 -= 2
	subfsr	FSR0, 0x2		; FSR0 -= 2

	movf	POSTINC1, W		; read tag(e1) ; FSR1 ++
	subwf	INDF0, W		; tag(e2) - tag(e1)
	bnc	caml_compare_CD_1gt2
	bnz	caml_compare_CD_1lt2

	movf	POSTINC0, W
	xorlw	STRING_TAG		; strings ?
	bz	caml_compare_D_string_compare
	xorlw	(STRING_TAG^DOUBLE_TAG) ; float ?
	bz	caml_compare_D_float_compare
	xorlw	(DOUBLE_TAG^DOUBLE_ARRAY_TAG) ; float[] ?
	bz	caml_compare_D_float_array_compare
	xorlw	(DOUBLE_ARRAY_TAG^ABSTRACT_TAG) ; abstract ?
	bz	caml_compare_CD_raise_ia_compare
	xorlw	(ABSTRACT_TAG^CLOSURE_TAG) ; closure ?
	bz	caml_compare_CD_raise_ia_compare
	xorlw	(CLOSURE_TAG^INFIX_TAG) ; infix ?
	bz	caml_compare_CD_raise_ia_compare
	
	movf	POSTINC1, W		; compare sizes
	subwf	INDF0, W		; size(e2) - size(e1)
	bnc	caml_compare_CD_1gt2
	bnz	caml_compare_CD_1lt2

;;; Complex case -> forward
	dcfsnz	TMP_REG_A, W		; if counter = 1
	bra	caml_compare_forward	; then skip push
caml_compare_push_forward:
	movlw	STACK_END		; check stack overflow
	cpfseq	FSR2H
	bra	caml_compare_canpush3
	movlw	-0x6
	addwf	FSR2L, W
	btfss	STATUS, C
	bra	caml_compare_raise_stack_overflow
caml_compare_canpush3:
	movf	TMP_REG_A, W		; stack[0] <- counter
	movwf	[0x1]
	movff	TMP_REG_C, POSTDEC2 	; push save(FSR0)
	movff	TMP_REG_B, POSTDEC2
	movff	TMP_REG_E, POSTDEC2 	; push save(FSR1)
	movff	TMP_REG_D, POSTDEC2
	subfsr	FSR2, 0x2	  	; push X
	bra	caml_compare_forward	; forward

caml_compare_C_compared:
	btfsc	ACCUL, 1		; sub-rountine returned 0 ?
	bra	caml_compare_CD_1ne2
	bra	caml_compare_C_1eq2
	
caml_compare_D_string_compare:
	rcall	caml_string_compare 	; compare string string
	bra	caml_compare_D_compared
caml_compare_D_float_compare:
	rcall	caml_float_compare 	; compare float float
	bra	caml_compare_D_compared
caml_compare_D_float_array_compare:
	rcall	caml_float_array_compare; compare float[] float[]
	bra	caml_compare_D_compared

caml_compare_D_compared:
	btfsc	ACCUL, 1		; sub-rountine returned 0 ?
	bra	caml_compare_CD_1ne2	; no -> finished
caml_compare_D_1eq2:
	movff	TMP_REG_B, FSR0L 	; restore FSR0 and FSR1
	movff	TMP_REG_C, FSR0H
	movff	TMP_REG_D, FSR1L
	movff	TMP_REG_E, FSR1H
caml_compare_C_1eq2:
	decfsz	TMP_REG_A, F		; decr counter ; counter = 0 ?
	bra	caml_compare_loop	; no -> loop
	movf	FSR2L, W		; FSR2 = save(FSR2) ?
	xorwf	TMP_REG_8, W
	bnz	caml_compare_goback
	movf	FSR2H, W
	xorwf	TMP_REG_9, W
	bz	caml_compare_CD_1eq2
caml_compare_goback:
	addfsr	FSR2, 0x2		; pop X
	movff	PREINC2, FSR1L		; pop FSR1
	movff	PREINC2, FSR1H
	movff	PREINC2, FSR0L		; pop FSR0
	movff	PREINC2, FSR0H
	decf	[0x1], W		; peek counter--
	movwf	TMP_REG_A
	bra	caml_compare_loop	; loop
#endif

;;;
	
#ifdef caml_useprim_caml_string_compare
caml_string_compare:
	;; ACCU = str1
	;; [0x2]:[0x1] = str2
	movff	FSR2L, TMP_REG_1	; save stack top
	movff	FSR2H, TMP_REG_2
	movsf	[0x1], TMP_REG_3	; FSR2 <- str2
	movsf	[0x2], TMP_REG_4
	movff	TMP_REG_3, FSR2L
	movff	TMP_REG_4, FSR2H
	movff	ACCUL, FSR0L		; FSR0 <- str1
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1		; FSR0 --
	subfsr	FSR2, 0x1		; FSR2 --
	decf	POSTINC0, W		; TMP_REG_{3,5} <- blk_size(str1) - 1
	movwf	TMP_REG_3
	movwf	TMP_REG_5
	decf	POSTINC2, W		; TMP_REG_4 <- blk_size(str2) - 1
	movwf	TMP_REG_4
	subwf	TMP_REG_5, W		; blk_size(str1) - blk_size(str2)
	bnc	caml_string_compare_L0
	movff	TMP_REG_4, TMP_REG_5	; TMP_REG_5 = min(blk_sizes) - 1
caml_string_compare_L0:
	movf	TMP_REG_5, F
	bz	caml_string_compare_loop_end
caml_string_compare_loop:
	movf	POSTINC0, W
	subwf	POSTINC2, W		; str2.[i] - str1.[i]
	bnz	caml_string_compare_diff
	movf	POSTINC0, W
	subwf	POSTINC2, W
	bnz	caml_string_compare_diff
	decfsz	TMP_REG_5, F		; loop min(blk_size) - 1 times
	bra	caml_string_compare_loop
caml_string_compare_loop_end:
	movf	TMP_REG_4, W
	subwf	TMP_REG_3, W		; blk_size(str1) - blk_size(str2)
	bz	caml_string_compare_blen_eq
	bnc	caml_string_compre_bl1_lt_bl2
caml_string_compare_bl1_gt_bl2:
	movf	[0x1], W
	bz	caml_string_compare_cont1
	bra	caml_string_compare_1gt2
caml_string_compre_bl1_lt_bl2:
	addfsr	FSR0, 0x1
	movf	POSTDEC0, W
	bz	caml_string_compare_cont1
	bra	caml_string_compare_1lt2
caml_string_compare_blen_eq:
	movf	[0x1], W
	bz	caml_string_compare_l2odd
	addfsr	FSR0, 0x1
	movf	POSTDEC0, W
	bz	caml_string_compare_1gt2
	M_CONST 0		 	; str1 = str2 ; return 0
	movff	TMP_REG_1, FSR2L 	; restore stack top
	movff	TMP_REG_2, FSR2H
	return
caml_string_compare_l2odd:
	addfsr	FSR0, 0x1
	movf	POSTDEC0, W
	bnz	caml_string_compare_1lt2
	movf	POSTINC0, W
	subwf	POSTINC2, W
	bnz	caml_string_compare_diff
	M_CONST 0		 	; str1 = str2 ; return 0
	movff	TMP_REG_1, FSR2L 	; restore stack top
	movff	TMP_REG_2, FSR2H
	return
caml_string_compare_cont1:
	movf	POSTINC0, W
	subwf	POSTINC2, W
caml_string_compare_diff:
	bc	caml_string_compare_1lt2
caml_string_compare_1gt2
	M_CONST	1		 	; str1 > str2 ; return 1
	movff	TMP_REG_1, FSR2L 	; restore stack top
	movff	TMP_REG_2, FSR2H
	return
caml_string_compare_1lt2:		; str1 < str2
	setf	ACCUL			; return -1
	setf	ACCUH
	movff	TMP_REG_1, FSR2L 	; restore stack top
	movff	TMP_REG_2, FSR2H
	return
#endif

#ifdef caml_useprim_caml_string_notequal
#ifndef caml_useprim_caml_string_equal
#define caml_useprim_caml_string_equal
#endif
caml_string_notequal:
	rcall	caml_string_equal
	btg	ACCUL, 1
	return
#endif

#ifdef caml_useprim_caml_string_equal
caml_string_equal:
	;; ACCU = str1
	;; [0x2]:[0x1] = str2
	movff	FSR2L, TMP_REG_1	; save stack top
	movff	FSR2H, TMP_REG_2
	movsf	[0x1], TMP_REG_3	; FSR2 <- str2
	movsf	[0x2], TMP_REG_4
	movff	TMP_REG_3, FSR2L
	movff	TMP_REG_4, FSR2H
	movff	ACCUL, FSR0L		; FSR0 <- str1
	movff	ACCUH, FSR0H
	subfsr	FSR0, 0x1		; FSR0 --
	subfsr	FSR2, 0x1		; FSR2 --
	movf	POSTINC0, W		; compare size
	xorwf	INDF2, W
	bnz	caml_string_equal_diff  ; if different sizes -> return false
	movff	POSTINC2, TMP_REG_3 	; TMP_REG_3 <- size
caml_string_equal_loop:
	movf	POSTINC0, W
	xorwf	POSTINC2, W
	bnz	caml_string_equal_diff
	movf	POSTINC0, W
	xorwf	POSTINC2, W
	bnz	caml_string_equal_diff
	decfsz	TMP_REG_3, F
	bra	caml_string_equal_loop
	M_CONST	1		 	; return 1 = true
	movff	TMP_REG_1, FSR2L 	; restore stack top
	movff	TMP_REG_2, FSR2H
	return
caml_string_equal_diff:
	M_CONST	0		 	; return 0 = false
	movff	TMP_REG_1, FSR2L 	; restore stack top
	movff	TMP_REG_2, FSR2H
	return
#endif

;;;

#ifdef caml_useprim_caml_int_compare
caml_int_compare:
	;; ACCU = v1
	;; [0x2]:[0x1] = v2
	movlw	0x80		; ACCUH = v1H + 0x80
	addwf	ACCUH, F
	addwf	[0x2], W	; W <- v2H + 0x80
	subwf	ACCUH, W	; (v1H + 0x80) - (v2H + 0x80)
	bnz	caml_int_compare_L0
	movf	[0x1], W
	subwf	ACCUL, W	; v1L - v2L
	bz	caml_int_compare_1eq2
caml_int_compare_L0:
	bnc	caml_int_compare_1lt2
	M_CONST	1		; v1 > v2 => return 1
	return
caml_int_compare_1lt2:
	setf	ACCUH		; v1 < v2 => return -1
	setf	ACCUL
	return
caml_int_compare_1eq2:
	M_CONST 0		; v1 = v2 => return 0
	return
#endif

;;;

#ifdef caml_useprim_caml_float_array_compare
#ifndef caml_useprim_caml_float_compare_1xx2
#define caml_useprim_caml_float_compare_1xx2
#endif
caml_float_array_compare:
	movff	FSR1L, TMP_REG_1 	; save FSR1
	movff	FSR1H, TMP_REG_2
	movff	ACCUL, FSR0L		; FSR0 <- @f2[]
	movff	ACCUH, FSR0H
	movsf	[0x1], FSR1L		; FSR1 <- @f1[]
	movsf	[0x2], FSR1H
	subfsr	FSR0, 0x1
	subfsr	FSR1, 0x1
	movf	INDF0, W
	subwf	POSTINC1, W		; size1 - size2
	bnc	caml_float_compare_1gt2
	bnz	caml_float_compare_1lt2
	movf	POSTINC0, W		; size = 0 ?
	bz	caml_float_compare_1eq2
	movwf	TMP_REG_4		; TMP_REG_4 <- size
	addfsr	FSR0, 0x1
	addfsr	FSR1, 0x1
caml_float_array_compare_loop:
	btfsc	INDF0, 7		; test sign
	bra	caml_float_array_compare_XN
	btfss	INDF1, 7
	bra	caml_float_array_compare_PP
	bra	caml_float_compare_1lt2
caml_float_array_compare_XN:
	btfss	INDF1, 7
	bra	caml_float_compare_1gt2
caml_float_array_compare_NN:
	movf	FSR1L, W		; FSR1 <-> FSR0
	movff	FSR0L, FSR1L
	movwf	FSR0L
	movf	FSR1H, W
	movff	FSR0H, FSR1H
	movwf	FSR0H
caml_float_array_compare_PP:
	movf	INDF1, W		; TMP_REG_3 <- e1 + 0x80
	andlw	B'00011111'
	btfsc	WREG, 4
	iorlw	B'11100000'
	addlw	0x80
	movwf	TMP_REG_3
	movf	INDF0, W		; W <- e2 + 0x80
	andlw	B'00011111'
	btfsc	WREG, 4
	iorlw	B'11100000'
	addlw	0x80
	subwf	TMP_REG_3, W		; (e1 + 0x80) - (e2 + 0x80)
	bnc	caml_float_compare_1gt2
	bnz	caml_float_compare_1lt2
	swapf	POSTDEC1, W		; TMP_REG_3 <- m1H
	rrncf	WREG, F
	andlw	B'00000011'
	movwf	TMP_REG_3
	swapf	POSTDEC0, W		; W <- m2H
	rrncf	WREG, F
	andlw	B'00000011'
	subwf	TMP_REG_3, W		; m1H - m2H
	bnc	caml_float_compare_1gt2
	bnz	caml_float_compare_1lt2
	movf	POSTINC0, W
	subwf	POSTINC1, W		; m1L - m2L
	bnc	caml_float_compare_1gt2
	bnz	caml_float_compare_1lt2
	dcfsnz	TMP_REG_4, F		; loop size times
	bra	caml_float_compare_1eq2
	btfss	INDF1, 7		; NN ?
	bra	caml_float_array_compare_continue
	movf	FSR1L, W		; FSR1 <-> FSR0
	movff	FSR0L, FSR1L
	movwf	FSR0L
	movf	FSR1H, W
	movff	FSR0H, FSR1H
	movwf	FSR0H
caml_float_array_compare_continue:
	addfsr	FSR0, 0x2
	addfsr	FSR1, 0x2
	bra	caml_float_array_compare_loop
#endif

#ifdef caml_useprim_caml_float_compare
#ifndef caml_useprim_caml_float_compare_1xx2
#define caml_useprim_caml_float_compare_1xx2
#endif
#endif
#ifdef caml_useprim_caml_float_compare_1xx2
caml_float_compare_1eq2:
	movff	TMP_REG_1, FSR1L 	; restore FSR1
	movff	TMP_REG_2, FSR1H
	M_CONST	0			; return 0
	return
caml_float_compare_1lt2:
	movff	TMP_REG_1, FSR1L 	; restore FSR1
	movff	TMP_REG_2, FSR1H
	setf	ACCUL
	setf	ACCUH
	return
caml_float_compare_1gt2:
	movff	TMP_REG_1, FSR1L 	; restore FSR1
	movff	TMP_REG_2, FSR1H
	M_CONST	1		 	; return 1
	return
#endif

#ifdef caml_useprim_caml_float_compare
#ifndef caml_useprim_caml_float_compare_1xx2
#define caml_useprim_caml_float_compare_1xx2
#endif
caml_float_compare:
	movff	FSR1L, TMP_REG_1 	; save FSR1
	movff	FSR1H, TMP_REG_2
	movff	ACCUL, FSR0L		; FSR0 <- @f2
	movff	ACCUH, FSR0H
	movsf	[0x1], FSR1L		; FSR1 <- @f1
	movsf	[0x2], FSR1H
	addfsr	FSR0, 0x1
	addfsr	FSR1, 0x1
	btfsc	INDF0, 7		; test sign
	bra	caml_float_compare_XN
	btfss	INDF1, 7
	bra	caml_float_compare_PP
	bra	caml_float_compare_1lt2
caml_float_compare_XN:
	btfss	INDF1, 7
	bra	caml_float_compare_1gt2
caml_float_compare_NN:
	movf	FSR1L, W		; FSR1 <-> FSR0
	movff	FSR0L, FSR1L
	movwf	FSR0L
	movf	FSR1H, W
	movff	FSR0H, FSR1H
	movwf	FSR0H
caml_float_compare_PP:
	movf	INDF1, W		; TMP_REG_3 <- e1 + 0x80
	andlw	B'00011111'
	btfsc	WREG, 4
	iorlw	B'11100000'
	addlw	0x80
	movwf	TMP_REG_3
	movf	INDF0, W		; W <- e2 + 0x80
	andlw	B'00011111'
	btfsc	WREG, 4
	iorlw	B'11100000'
	addlw	0x80
	subwf	TMP_REG_3, W		; (e1 + 0x80) - (e2 + 0x80)
	bnc	caml_float_compare_1gt2
	bnz	caml_float_compare_1lt2
	swapf	POSTDEC1, W		; TMP_REG_3 <- m1H
	rrncf	WREG, F
	andlw	B'00000011'
	movwf	TMP_REG_3
	swapf	POSTDEC0, W		; W <- m2H
	rrncf	WREG, F
	andlw	B'00000011'
	subwf	TMP_REG_3, W		; m1H - m2H
	bnc	caml_float_compare_1gt2
	bnz	caml_float_compare_1lt2
	movf	INDF0, W
	subwf	INDF1, W		; m1L - m2L
	bnc	caml_float_compare_1gt2
	bnz	caml_float_compare_1lt2
	;; fallthrough
#endif

;;;

#ifdef caml_useprim_caml_compare_to_eq
caml_compare_to_eq:
	btg	ACCUL, 1		; 00 -> 1
	movlw	0x03			; 01 -> 0
	andwf	ACCUL, F		; 11 -> 0
	clrf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_compare_to_neq
caml_compare_to_neq:
	movlw	0x03			; 00 -> 0
	andwf	ACCUL, F		; 01 -> 1
	clrf	ACCUH			; 11 -> 1
	return
#endif

#ifdef caml_useprim_caml_compare_to_lt
caml_compare_to_lt:
	rrncf	ACCUL, F		; 00 -> 0
	bsf	ACCUL, 0		; 01 -> 0
	movlw	0x03			; 11 -> 1
	andwf	ACCUL, F
	clrf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_compare_to_le
caml_compare_to_le:
	btg	ACCUL, 1		; 00 -> 1
	btfsc	ACCUL, 2		; 01 -> 0
	bsf	ACCUL, 1		; 11 -> 1
	movlw	0x03
	andwf	ACCUL, F
	clrf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_compare_to_gt
caml_compare_to_gt:
	btfsc	ACCUL, 2		; 00 -> 0
	bcf	ACCUL, 1		; 01 -> 1
	movlw	0x03			; 11 -> 0
	andwf	ACCUL, F
	clrf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_compare_to_ge
caml_compare_to_ge:
	rrncf	ACCUL, F		; 00 -> 1
	btg	ACCUL, 1		; 01 -> 1
	bsf	ACCUL, 0		; 11 -> 0
	movlw	0x03
	andwf	ACCUL, F
	clrf	ACCUH
	return
#endif

#ifdef caml_useprim_caml_raise_ia_compare
#ifndef caml_useprim_caml_raise_ia
#define caml_useprim_caml_raise_ia
#endif
caml_raise_ia_compare:
	movlw	0x1			; erase ACCU and stack[0]
	movwf	ACCUL
	movwf	[0x1]
	clrf	ACCUH
	clrf	[0x2]
	M_CHECK_UNFULL_HEAP 0x7		; check allocation of string and exn
	M_WRITE_BYTE STRING_TAG		; write string tag
	M_WRITE_BYTE 0x4		; write string size
	movff	FSR1L, TMP_REG_1	; mem string addr
	movff	FSR1H, TMP_REG_2
	M_WRITE_BYTE 'c'		; write string content
	M_WRITE_BYTE 'o'
	M_WRITE_BYTE 'm'
	M_WRITE_BYTE 'p'
	M_WRITE_BYTE 'a'
	M_WRITE_BYTE 'r'
	M_WRITE_BYTE 'e'
	M_WRITE_BYTE 0x0		; write string \0
	goto	caml_raise_ia
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;               TOOL                ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifdef caml_useprim_caml_raise_ia
caml_raise_ia:
	M_WRITE_BYTE 0x0		; write exn tag
	M_WRITE_BYTE 0x2		; write exn size
	movff	FSR1L, ACCUL		; ACCU <- exn
	movff	FSR1H, ACCUH
	lfsr	FSR0, (STACK_ANCHOR - 0x1 - (0x2 * INVALID_ARG_IND))
	movff	POSTINC0, POSTINC1 	; write global(INVALID_ARG_IND)
	movff	POSTINC0, POSTINC1
	movff	TMP_REG_1, POSTINC1 	; write string addr
	movff	TMP_REG_2, POSTINC1
	goto	_RAISE
#endif


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;           UNIMPLEMENTED           ;;;;;;;;
;;;;;;;;                                   ;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   caml_sub_float:
;   caml_mul_float:
;   caml_div_float:
;   caml_exp_float:
;   caml_floor_float:
;   caml_fmod_float:
;   caml_frexp_float:
;   caml_ldexp_float:
;   caml_log_float:
;   caml_log10_float:
;   caml_modf_float:
;   caml_sqrt_float:
;   caml_power_float:
;   caml_sin_float:
;   caml_sinh_float:
;   caml_cos_float:
;   caml_cosh_float:
;   caml_tan_float:
;   caml_tanh_float:
;   caml_asin_float:
;   caml_acos_float:
;   caml_atan_float:
;   caml_atan2_float:
;   caml_ceil_float:
;   caml_classify_float:
;   caml_int_of_float:
;   caml_float_of_int:

;   caml_int32_neg:
;   caml_int32_add:
;   caml_int32_sub:
;   caml_int32_mul:
;   caml_int32_div:
;   caml_int32_mod:
;   caml_int32_and:
;   caml_int32_or:
;   caml_int32_xor:
;   caml_int32_shift_left:
;   caml_int32_shift_right:
;   caml_int32_shift_right_unsigned:
;   caml_int32_of_int:
;   caml_int32_to_int:
;   caml_int32_of_float:
;   caml_int32_to_float:
;   caml_int32_compare:
;   caml_int32_format:
;   caml_int32_of_string:
;   caml_int32_bits_of_float:
;   caml_int32_float_of_bits:

;   caml_int64_neg:
;   caml_int64_add:
;   caml_int64_sub:
;   caml_int64_mul:
;   caml_int64_div:
;   caml_int64_mod:
;   caml_int64_and:
;   caml_int64_or:
;   caml_int64_xor:
;   caml_int64_shift_left:
;   caml_int64_shift_right:
;   caml_int64_shift_right_unsigned:
;   caml_int64_of_int:
;   caml_int64_to_int:
;   caml_int64_of_float:
;   caml_int64_to_float:
;   caml_int64_of_int32:
;   caml_int64_to_int32:
;   caml_int64_of_nativeint:
;   caml_int64_to_nativeint:
;   caml_int64_compare:
;   caml_int64_format:
;   caml_int64_of_string:
;   caml_int64_bits_of_float:
;   caml_int64_float_of_bits:

;   caml_nativeint_neg:
;   caml_nativeint_add:
;   caml_nativeint_sub:
;   caml_nativeint_mul:
;   caml_nativeint_div:
;   caml_nativeint_mod:
;   caml_nativeint_and:
;   caml_nativeint_or:
;   caml_nativeint_xor:
;   caml_nativeint_shift_left:
;   caml_nativeint_shift_right:
;   caml_nativeint_shift_right_unsigned:
;   caml_nativeint_of_int:
;   caml_nativeint_to_int:
;   caml_nativeint_of_float:
;   caml_nativeint_to_float:
;   caml_nativeint_of_int32:
;   caml_nativeint_to_int32:
;   caml_nativeint_compare:
;   caml_nativeint_format:
;   caml_nativeint_of_string:

;   caml_lazy_follow_forward:
;   caml_lazy_make_forward:

;   caml_weak_create:
;   caml_weak_set:
;   caml_weak_get:
;   caml_weak_get_copy:
;   caml_weak_check:
;   caml_weak_blit:

;   caml_get_public_method:

;;; 

;   caml_gc_stat:
;   caml_gc_quick_stat:
;   caml_gc_counters:
;   caml_gc_get:
;   caml_gc_set:
;   caml_gc_minor:
;   caml_gc_major:
;   caml_gc_full_major:
;   caml_gc_major_slice:
;   caml_gc_compaction:

;   caml_ml_open_descriptor_in:
;   caml_ml_open_descriptor_out:
;   caml_ml_out_channels_list:
;   caml_channel_descriptor:
;   caml_ml_close_channel:
;   caml_ml_channel_size:
;   caml_ml_channel_size_64:
;   caml_ml_set_binary_mode:
;   caml_ml_flush_partial:
;   caml_ml_flush:
;   caml_ml_output_char:
;   caml_ml_output_int:
;   caml_ml_output_partial:
;   caml_ml_output:
;   caml_ml_seek_out:
;   caml_ml_seek_out_64:
;   caml_ml_pos_out:
;   caml_ml_pos_out_64:
;   caml_ml_input_char:
;   caml_ml_input_int:
;   caml_ml_input:
;   caml_ml_seek_in:
;   caml_ml_seek_in_64:
;   caml_ml_pos_in:
;   caml_ml_pos_in_64:

;   caml_output_value:
;   caml_output_value_to_string:
;   caml_output_value_to_buffer:
;   caml_input_value:
;   caml_input_value_from_string:
;   caml_marshal_data_size:

;   caml_ml_input_scan_line:
;   caml_lex_engine:
;   caml_new_lex_engine:
;   caml_parse_engine:

;   caml_sys_exit:
;   caml_sys_open:
;   caml_sys_close:
;   caml_sys_file_exists:
;   caml_sys_is_directory:
;   caml_sys_remove:
;   caml_sys_rename:
;   caml_sys_chdir:
;   caml_sys_getcwd:
;   caml_sys_getenv:
;   caml_sys_get_argv:
;   caml_sys_system_command:
;   caml_sys_time:
;   caml_sys_random_seed:
;   caml_sys_get_config:
;   caml_sys_read_directory:

;   caml_terminfo_setup:
;   caml_terminfo_backup:
;   caml_terminfo_standout:
;   caml_terminfo_resume:

;   caml_md5_string:
;   caml_md5_chan:

;   caml_dynlink_open_lib:
;   caml_dynlink_close_lib:
;   caml_dynlink_lookup_symbol:
;   caml_dynlink_add_primitive:
;   caml_dynlink_get_current_libs:

;   caml_get_global_data:
;   caml_get_section_table:
;   caml_reify_bytecode:
;   caml_realloc_global:
;   caml_get_current_environment:
;   caml_invoke_traced_function:
;   caml_static_alloc:
;   caml_static_free:
;   caml_static_release_bytecode:
;   caml_static_resize:
;   caml_install_signal_handler:
;   caml_bitvect_test:
;   caml_register_named_value:
;   caml_final_register:
;   caml_final_release:
;   caml_ensure_stack_capacity:
