(*************************************************************************)
(*                                                                       *)
(*                                OCaPIC                                 *)
(*                                                                       *)
(*            Benoit Vaugon, Université Pierre et Marie Curie            *)
(*                                                                       *)
(*    Ce fichier est distribué sous les termes de la licence CeCILL-B    *)
(*    décrite dans le fichier ../../../LICENCE.                          *)
(*                                                                       *)
(*************************************************************************)

open Printf;;
open Graphics;;

exception Exn of string

let green = rgb 0 200 50;;
let blue = rgb 0 50 200;;

let column_nb = ref (-1)
let line_nb = ref (-1)

let sync_display =
  let sync_mutex = Mutex.create () in
  let sync_check = ref false in
  let rec sync_thread () =
    Mutex.lock sync_mutex;
    let must_sync = !sync_check in
    sync_check := false;
    Mutex.unlock sync_mutex;
    if must_sync then synchronize ();
    Thread.delay 0.1;
    sync_thread ();
  in
  let _ = Thread.create sync_thread () in
    fun () ->
      Mutex.lock sync_mutex;
      sync_check := true;
      Mutex.unlock sync_mutex;
;;

let display_border () =
  let width = size_x () in
  let height = size_y () in
    set_color green;
    fill_rect 0 0 width height;
    set_color black;
    fill_rect 10 10 (width - 20) (height - 20);
    set_color blue;
    fill_rect 20 20 (width - 40) (height - 40);
;;

let init_display column_nb_init line_nb_init =
  let win_width = 65 + 23 * column_nb_init in
  let win_height = 65 + 34 * line_nb_init in
  let config = Printf.sprintf " %dx%d" win_width win_height in
  let error () =
    raise (Exn (sprintf "invalid size %dx%d" column_nb_init line_nb_init));
  in
    begin match line_nb_init with
      | 1 | 2 | 4 ->
	  if column_nb_init <= 0 || column_nb_init * line_nb_init > 80 then
	    error ()
      | _ -> error ()
    end;
    column_nb := column_nb_init;
    line_nb := line_nb_init;
    open_graph config;
    set_window_title "LCD display";
    display_mode false;
    display_border ();
    sync_display ();
;;

let display_char x y =
  let disp i j c =
    set_color (if c then white else blue);
    fill_rect (x + 4 * i) (y + 4 * (7 - j)) 2 2;
  in
    Array.iteri (fun j -> Array.iteri (fun i c -> disp i j c));
;;

let show matrix =
  let height = size_y () in
  let disp i j c = display_char (36 + 23 * i) (height - 65 - 34 * j) c in
    display_border ();
    begin match !line_nb with
      | 1 ->
	  for i = 0 to pred !column_nb do
	    disp i 0 matrix.(i).(0)
	  done
      | 2 ->
	  for i = 0 to pred !column_nb do
	    disp i 0 matrix.(i).(0);
	    disp i 1 matrix.(i).(1);
	  done
      | 4 ->
	  for i = 0 to pred !column_nb do
	    disp i 0 matrix.(i).(0);
	    disp i 1 matrix.(i).(1);
	    disp i 2 matrix.(i + 20).(0);
	    disp i 3 matrix.(i + 20).(1);
	  done
      | _ -> ()
    end;
    sync_display ();
;;
